
#ifndef GRIDCELL
#define GRIDCELL

#include "SoccerTypes.h"
#include "Geometry.h"     // needed for AngDeg
#include "math.h"       // needed for M_PI constant


/******************************************************************************/
/**********************  CLASS GRIDCELL   *************************************/
/******************************************************************************/
// This class contains info about cells of the grid used for making decisions
// about actions with the ball. Each cell is a sector. Created by Vadim 10/06/03
class GridCell
{
public:
  bool         isActive;          // true if the cell is in use

  // *** cell geomentry parameters ***
  GridSectorT  sectorType;
  VecPosition globalCenter;       // cell center in global Cartesian coordinates
  VecPosition relativeCenter;     // cell center in relative Cartesian coordinates
  VecPosition gridCenter;         // grid center in global Cartesian coordinates
  AngDeg centerDir;               // cell center direction from the grid center
  double centerDist;              // cell center distance from the grid center
  double cellArea;                // cell area
  double cellRadius;              // cell efective radius

  ObjectT oFastestTeammate;       // fastest-teammate-to-the-cell-center object type

  // ***  criteria        
  double dImportance;              // cell importance factor in 0..100 range
  double dExpectedOppFactor;       // ~density of opponents potentially expected in the cell
  double dExpectedTmmFactor;       // ~density of teammates potentially expected in the cell
  int iCyclesBallToReach;          // number of cycles to deliver ball to the cell center
  int iCyclesFastestOpponent;      // number of cycles to reach the cell center by the fastest opponent
  int iCyclesFastestTeammate;      // number of cycles to reach the cell center by the fastest teammate

public:

  GridCell ( );                    // default constructor
  void        initCell                ( VecPosition             agPos,
                                        AngDeg                  agDir,
                                        double                  ringRadius,
                                        AngDeg                  ang,
                                        double                  area                    );
  AngDeg      getCellDir              (                                                 );  
  double      getCellDist             (                                                 );
  void        eliminate               (                                                 );
  bool        isInUse                 (                                                 );
  VecPosition getCellCenter           (                                                 );
  VecPosition getGridCenter           (                                                 );
  double      getCellEffRadius        (                                                 );
  double      getCellArea             (                                                 );
  double      getTmmFactor            (                                                 );
  void        setTmmFactor            ( double                  num                     );
  double      getOppFactor            (                                                 );
  void        setOppFactor            ( double                  num                     );
  void        setImportance           ( double                  val                     );
  double      getImportance           (                                                 );
  double      calcPresenseFactor      ( VecPosition             pos,
                                        bool                    isTeammate              );
  void        setCyclesBallToReach    ( int                     cyc                     );
  int         getCyclesBallToReach    (                                                 );
  void        setCyclesFastestTmm     ( int                     cyc                     );
  int         getCyclesFastestTmm     (                                                 );
  void        setCyclesFastestOpp     ( int                     cyc                     );
  int         getCyclesFastestOpp     (                                                 );
  void        setFastestTeammate      ( ObjectT                 tmm                     );
  ObjectT     getFastestTeammate      (                                                 );
  GridSectorT getSectorType           (                                                 );

  
} ;  // class GridCell



#endif
