/*
Copyright (c) 2000-2002, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the name of the University of Amsterdam nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*! \file Player.h
<pre>
<b>File:</b>          Player.h
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       03/03/2001
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This file contains the declaration for the Player class,
               which is a superclass from BasicPlayer and contains the
               decision procedure to select the skills from the
               BasicPlayer. 
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
03/03/2001       Jelle Kok       Initial version created
</pre>
*/

#ifndef PLAYER
#define PLAYER

#include "BasicPlayer.h"
#include "Formations.h"     // needed for Formations
#include "level.h"
#include "GridCell.h"

class Level1;
class Level2;
class Level3;
class Level4;

void* stdin_callback( void * v );


/*! This class is a superclass from BasicPlayer and contains a more
    sophisticated decision procedure to determine the next action. */
    
class Player:public BasicPlayer
{
  Formations    *formations;             /*!< all formation information      */
  bool          bContLoop;               /*!< is server is alive             */

  Time          m_timeLastSay;           /*!< last time communicated         */
  ObjectT       opponentToMark;          /*!< the current opponent to mark   */

  /* !*this is a section of variables for using levels structures. These values the agent
   must "remember" to use the level structure*/

  int           currentLevel;            /*!< the number of the current level*/
     
  //variables of the level 3
  //bool          ifWePossessBall;       // indicates if our team posesses the ball
  /*! this parameter has been moved into level 3*/
  // the other object of this level is the targrt area, but it is implemented as a part
  // of the WorldModel
  
  //  variables of the level 2
  bool          ifIPossessBall  ;     // indicates if the player posesses the ball
  ActionT       currentAction   ;     // the action for level 1
  ObjectT       teammateToPass  ;     // the current teammate to pass in level 1
  VecPosition   pointToPass     ;     // the current point to pass in level 1
  AngDeg        dribbleDirection;     // the direction for dribbling for level 1
  DribbleT      dribbleType     ;     // type of dribbling for level 1
  ObjectT       opponentToOutplay;
  VecPosition   shootingPoint;
  double        shootingConeWidth;

  int CycleCounter; //temporary variable for debugging
  int iSayBallCount;                    // Say-Ball message counter -- added by Vadim   
  AngDeg        ballTurnIncrement;    // used by trickOppMultyCycle()  
  
  // variables of the level 1
  int  multiCycleDribbleStep;            // indicates what step of multicycle command is the current
  int  cycleMultiCounterLastUpdated;     // tells when the last update of multiCycleDribbleStep was

  // attribiutes associated with decision making whilst acting with the ball
  GridCell gridcells[ MAX_GRIDCELLS ];   // gridcell storage
  
  vector<Level*> levels;                  /*!< Action levels for player architecture */

  void          updateAllTimers           (                                   );
  int           nextMultyCycleCount       (                                   );
        
  void          goalieMainLoop            (                                   );
  void          defenderMainLoop          (                                   );
  void          midfielderMainLoop        (                                   );
  void          attackerMainLoop          (                                   );

  
  // methods associated with saying (defined in Player.C)
  bool          shallISaySomething        (                                   );
  bool          amIAgentToSaySomething    (                                   );
  SoccerCommand sayBallStatus             (                                   );
  SoccerCommand sayTargetArea             ( int nextCoordinator,
                                            int duration,
                                            VecPosition targetAreaCentre,
                                            bool offence                      );

   // added by Vadim
  void printStatistics                    (                                   );
  void sendSayBallMsg                     (                                   );
  SoccerCommand sayBallStatus             ( int ballActionType,
                                            int playerID,        
                                            int nr_cyclces,      
                                            double confidence,
                                            bool lowSta,         
                                            bool retransmit                   );
  bool          shouldIRequestForPassing  (                                   );
  bool          shouldGoalieRequest       ( VecPosition    *pointToPass,
                                            int            *iCycles           );

  double        getConfidenceSayBall      (                                   );                                                                                                                               

  // *** methods associated with decision making whilst acting with the ball
  bool initGrid                           ( int            *numOfAlternatives );
  bool eliminateGridCells                 ( double          maxDist,
                                            int             *count            );
  void evaluateCongestion                 ( int             *eliminatedCount  );
  bool eliminateCongestedCell             ( int             idx               );
  void printGridCells                     (                                   );
  void evaluateImportance                 (                                   );
  int  addMoreImportantCells              ( int              count            );
  void calcGridTimeBalance                ( int             *count            );
  double  evaluatePresense                ( int              idx,
                                            ObjectSetT       set              );
  double getPassRequestFactor             ( VecPosition      pos              );
  double calcPosImportance                ( VecPosition      pos              );
  double calcImportanceMiddle             ( VecPosition      pos              );
  double calcImportanceOffense            ( VecPosition      pos              );
  double calcImportanceDefense            ( VecPosition      pos              );
  void   putGridStatistics                (                                   );
  bool   optimizeActionWithTheBall        ( int              numOfAlt         );
  bool canImoveBallWell                   ( ObjectT         *tmate,
                                            VecPosition     *pos,
                                            int             *iCycles          );


  //Action selection methods
  SoccerCommand UVA2001Goalie             (                                   );
  ActionT       determineActionMode       (                                   );
  ActionT       determineGoalieActionMode (                                   );
  bool          generateAction            (                                   );

  
  bool          generateGoalieAction      (                                   );
  SoccerCommand determineActionWithBall   (                                   );
  SoccerCommand adjustDashPowerForStamina ( SoccerCommand action,
                                            ActionT mode,
                                            bool isGoalie = false             );
  SoccerCommand goalieInterceptOnLine     (                                   );
  SoccerCommand goalieInterceptCrossPass  (                                   );
  ObjectT       getMostFreeForwardTeammate(                                   );
  ObjectT       getMostFreeForwardTeammateP(VecPosition point                 );
  ObjectT       getMostFree456Teammate    (                                   );
  ObjectT       getMostFreeRectTeammate   ( Rectangle rect                    );
  void          mapFreeTeammates          (ObjectT *leftBack,ObjectT *left,
                                           ObjectT *leftForward, ObjectT *forward,
                                           ObjectT *rightForward,
                                           ObjectT *right, ObjectT *rightBack,
                                           ObjectT *back,
                                           ObjectT *reg6,
                                           ObjectT *reg6a,
                                           VecPosition pos);
  
  bool          shouldIMark               (                                   );
  // added by S.Razykov 
  bool          shouldIMark1              ( VecPosition pos                   );
  
  bool          shouldIIntercept          ( InterceptionT *iType              );
  bool          shouldIHideBall           ( ObjectT                o          );
  bool          shouldITrickOpponent      ( ObjectT                o          );
  bool          shouldITrickGoalie        (                                   );
  bool          shouldIScore              ( VecPosition *pos                  );
  bool          shouldIScore              ( VecPosition *pos,
                                            ActionT     *action               );                  
  bool         shouldIScoreInSpiteOfGoalie( VecPosition *pos,
                                            ActionT     *action               );
  bool         shouldIShootNarrowly       ( AngDeg       anglWidth,
                                            VecPosition  leftCorner,
                                            VecPosition  rightCorner,
                                            VecPosition *pos,
                                            ActionT *action                   );
  double        getBestLeadingPass        ( ObjectT *o                        );
  bool          isItBreakIn               ( ObjectT teammate, ObjectT opponent,
                                            VecPosition posBall);
  
  double        calculatePriority         ( const VecPosition& opponentPos    );
  bool          amFastestToSideClearedBall(                                   );
  void          setUpForKickOff           (                                   );
  VecPosition   selectBestShootingPoint   (                                   );
  
  /* methods added for the using  of the level structure*/
  
  ObjectT       getMostFreeTargetTeammate (                                   );
  ObjectT       getMostFreeAroundTeammate ( double radius                     );

  LevelStateT   generateAction1           (                                   ); // for level 1
  ActionT       determineActionWithBall1  (                                   );
  ActionT       determineActionWithBall2  (                                   );
   
  ActionT       determineActionMode1      (                                   );

  // set variables for level structure methods
  //void          setTeamPossession         (                                   );
  void          setPlayerPossesion       (                                   );
  bool          getPlayerPossesion       (                                    );
  void          setCurrentAction          (                                   );

   // multicycles skills
  LevelStateT   dribbleSimpleMultiCycle   ( AngDeg ang,  DribbleT d,
                                            VecPosition  p                    );
  LevelStateT   outplaySimpleMultiCycle   ( VecPosition pos                   );
  LevelStateT   interceptSimpleMultiCycle ( bool isGoalie                     );
  LevelStateT   zoningDefenseMultiCycle  ( VecPosition center, bool isFree    );
  LevelStateT   zblokOpponentMultiCycle  ( ObjectT o,double dist,
                                                        double interval       );
  LevelStateT   anticipateBallMultiCycle  ( double dist                       );
  LevelStateT   directPassMultiCycle      ( ObjectT                o,
                                            PassT                passType     );
  LevelStateT   directPassMultiCycle      ( VecPosition pos,
                                            int                    TmmNr,
                                            PassT                passType     );
  LevelStateT   leadingPassMultiCycle      ( ObjectT                o         );
  LevelStateT   hideBallMultiCycle         ( ObjectT                o         );
  LevelStateT   holdBallMultiCycle         ( ObjectT                o         );
  LevelStateT   trickOppMultyCycle         ( ObjectT                o         );
                                                        
  
public:
  Player                                  ( ActHandler     *a,
                                            WorldModel     *wm,
                                            ServerSettings *ss,
                                            PlayerSettings *cs,
                                            Formations     *fs,
                                            char           *strTeamName,
                                            double         dVersion,
                                            int            iReconnect = -1    );

  void          mainLoop                  (                                   );
  void          brokenGoalieVersion1      (                                   );
  void          brokenPlayerVersion1      (                                   );

  
  // methods that deal with user input (from keyboard) to sent commands
  void          handleStdin               (                                   );
  void          showStringCommands        ( ostream& out                      );
  bool          executeStringCommand      ( char *str                         );

  // test methods
  void          test_only_update          (                                   );


  friend class Level1;  
  friend class Level2;
  friend class Level3;
  friend class Level4;
} ;


#endif
