/*
Copyright (c) 2000-2002, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the name of the University of Amsterdam nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*! \file Player.C

<pre>
<b>File:</b>          Player.C
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       03/03/2001
<b>Last Revision:</b> $ID$
<b>Contents:</b>      This file contains the definitions for the Player class,
               which is a superclass from BasicPlayer and contains the
               decision procedure to select the skills from the
               BasicPlayer. 
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
03/03/2001       Jelle Kok       Initial version created
</pre>
*/
#include "Player.h"
#include "Parse.h"

#include "level1.h"
#include "level2.h"
#include "level3.h"
#include "level4.h"

#include <sys/poll.h> // needed for 'poll'


/*!This is the constructor the Player class and calls the constructor of the
   superclass BasicPlayer.
   \param act ActHandler to which the actions can be sent
   \param wm WorldModel which information is used to determine action
   \param ss ServerSettings that contain parameters used by the server
   \param ps PlayerSettings that contain parameters important for the client
   \param strTeamName team name of this player
   \param dVersion version this basicplayer corresponds to
   \param iReconnect integer that defines player number (-1 when new player) */
Player::Player( ActHandler* act, WorldModel *wm, ServerSettings *ss,
      PlayerSettings *ps,
      Formations *fs, char* strTeamName, double dVersion, int iReconnect )

{
  char str[MAX_MSG];

  ACT           = act;
  WM            = wm;
  SS            = ss;
  PS            = ps;
  formations    = fs;
  bContLoop     = true;
  opponentToMark = OBJECT_ILLEGAL;
  
  WM->setTeamName( strTeamName );
  m_timeLastSay = -5;

  // Vadim begin
  iSayBallCount = 0;
// Vadim end
  
  //Level0* myLevel0 = new Level0(this);
  shootingConeWidth = 0.25;

  levels.push_back(new Level1(this));
  levels.push_back(new Level2(this));
  levels.push_back(new Level3(this));
  levels.push_back(new Level4(this));

  for (int i = 0; i < (int)levels.size()-1; ++i)
  {
    levels[i]->setHigherLevel(levels[i+1]);
  }

  for (int i = (int)levels.size()-1; i > 0; --i)
  {
    levels[i]->setLowerLevel(levels[i-1]);
  }
                                                                            
  currentLevel = levels.size()-1;
  
  /* !*this is a section of variables for using levels structures. These values the agent
   must "remember" to use the level structure*/

  //variables of the level 3
  // indicates if our team posesses the ball
  //ifWePossessBall = false; moved to lewel 3
  
  // variables of the level 2
  ifIPossessBall  = false;
  teammateToPass  = OBJECT_ILLEGAL;
  currentAction   = ACT_ILLEGAL;
  pointToPass.setVecPosition(0.0,0.0);
    
  // variables of the level 1   
  // this variable points to status of dribble the 0 is an initial value
  multiCycleDribbleStep = 0;
  
  // wait longer as role number increases, to make sure players appear at the
  // field in the correct order
  poll( 0, 0, formations->getPlayerInFormation()*100 );

  // create initialisation string
  if( iReconnect != -1 )
    sprintf( str, "(reconnect %s %d)", strTeamName, iReconnect );
  else if( formations->getPlayerType() == PT_GOALKEEPER )
    sprintf( str, "(init %s (version %f) (goalie))", strTeamName, dVersion );
  else
    sprintf( str, "(init %s (version %f))", strTeamName, dVersion );
  ACT->sendMessage( str );
}

/*! This is the main loop of the agent. This method calls the update methods
    of the world model after it is indicated that new information has arrived.
    After this, the correct main loop of the player type is called, which
    puts the best soccer command in the queue of the ActHandler. */
void Player::mainLoop( )
{
  while( bContLoop )                                 // as long as server alive
  {
    Log.logWithTime( 3, "  start update_all" );
    Log.setHeader( WM->getCurrentCycle(), WM->getPlayerNumber() );

   // if (WM->getCurrentCycle() == 0 && WM->getPlayerNumber() == 6)
   // {
     // ACT->sendCommandDirect(SoccerCommand(CMD_CHANGEVIEW,VA_WIDE, VQ_HIGH) );
   // }
    
    if( WM->updateAll( ) == true )
    {

      //It should be noted that the only level called from this control
      //loop is the currently selected one.  All responsibility for calling
      //lower levels is devolved to the levels in question. If a level needs to return
      //control to a higher level at any point, it should return indicating it wishes
      //to give up control, in which case the next highest level will be assigned control
      //and called.  Levels are responsible for their own timer maintenance, which assumes
      //that all levels will be well behaved.  However, if a low level is taking an
      //inordinate amount of time it can still be preempted by a higher level.

      //If this player isn't the coordinator and isn't paying attention to the current coordinator
      //send a new attentionto message to pay attention to the correct player.
      if (WM->getCurrentCycle() == 0 || (WM->getPlayerNumber() != WM->getCoordinatorNumber() && WM->getCoordinatorNumber() != WM->getCurrentAttentionPlayer()))
      { ACT->sendCommandDirect( SoccerCommand(CMD_ATTENTIONTO, 0, WM->getCoordinatorNumber(), 0));
        //{ cout << "Player " << WM->getPlayerNumber() << " send attention to with return true\n";
        //}
        WM->setCurrentAttentionPlayer(WM->getCoordinatorNumber());
      }
      else if (WM->getPlayerNumber() == WM->getCoordinatorNumber() && WM->getCurrentAttentionPlayer() != -1)
      { //ACT->sendCommandDirect( SoccerCommand(CMD_ATTENTIONTO, -1, -1, 0)); //turn off attention to
        //WM->setCurrentAttentionPlayer(-1);
      }
      
      updateAllTimers();
      int status = 6; //6 is an unassigned value
      //0 indicates actions selected and in process
      //1 indicates finished, no actions selected.
      //-1 indicates actions not selected (failure)
      int highestTimedOutLevel = -1;
      int numberOfLevels = (int)levels.size();
      int i = 0;
      //int j = 0;
      //while (status != 0/*actions not selected*/)  //how do we tell if there are actions selected?
      //This initial design requires that this loop/method know about all of the levels
      //that need to be executed, at least enough to call them, and must be able to tell
      //them to update their timer.
      while (status > 1/*actions not selected*/) 
      {
       // j++;
      //  if (j>2)
       // {
       //   cout<<"looping!!! "<<WM->getPlayerNumber()<<" j="<<j<<endl;
      //  }  
        highestTimedOutLevel = -1;
        //figure out the highest level that has just timed out (if any)
        //for (i = 0; i < numberOfLevels; ++i)
        for (i = currentLevel; i < numberOfLevels; ++i) // levels from current only
        { if (levels[i]->getTimer() == 0)
          { highestTimedOutLevel = i;
          }
        }
        //if there is a timed out level give control to the level above it, unless it is
        //the top level, in which case it gets control.
        if (highestTimedOutLevel >= 0)
        //{ if (highestTimedOutLevel < numberOfLevels-1)
        { if (highestTimedOutLevel>=currentLevel)
          { if (highestTimedOutLevel < numberOfLevels-1)
        
            { currentLevel = highestTimedOutLevel + 1;
            }
            else
            { currentLevel = highestTimedOutLevel;
             
            }
          }
          //if there was a timed out level expire all the levels below it.
          
          //for (j = 0; j < currentLevel; ++j)  PPS. I think we don't need to do it (S.R.)
          //{ levels[j]->expireTimer();
          //}
        }

      //if any level has timed out, give control to the level one higher
      //than it.
//        if (false/*any timers have expired*/)
//        {
          //make the level above the highest level with an expired timer the current level.
//        }
//        else if(true/*the current level has given up control*/)
//        {
          //make the level above the current level the current level.
//        }
        
        //call the main method of the current level //begins action selection
        //cout<<"ent level"<<currentLevel;
        status = levels[currentLevel]->execute();
        //cout << " Called Level " <<currentLevel;
        //cout<<" Status "<<status<<" "<<WM->getPlayerNumber()<<endl;

        if (status > 0) // it means the current level has finished its work (successfully or not)
         currentLevel++; // gives the control to higher level

         //cout<<"high lev "<<currentLevel;
              
      } //while
      //cout<<"Fin level "<<currentLevel<<endl;

      
      
      //after the levels have been called any and all actions should be selected already
      //so we just need to wait until the appropriate time to send the actions, and then
      //wait on new input.

 
      //**** generate a Say-Ball message if needed ****
      sendSayBallMsg();

      // set Team possesion with a one-cycle delay after the ball possesion has been
      // determined. If this delay is critical, use ball possesion staus.
      WM->setTeamPossesion (  WM->getBallPossesion() );

      Log.logWithTime( 3, "  determined action; waiting for new info" );

      // directly after see message, will nog get better info, so send commands
      if( WM->getTimeLastSeeMessage() == WM->getCurrentTime() )
        ACT->sendCommands( );
    }
    else
      Log.logWithTime( 3, "  HOLE no action determined; waiting for new info" );

    // wait for new information from the server
    // cannot say bContLoop=WM->wait... since bContLoop can be changed elsewhere
    if(  WM->waitForNewInformation() == false )
        bContLoop =  false;
  }

  // shutdow, print hole and number of players seen statistics
  printStatistics( );

} // end mainLoop


/*! This is the main decision loop for the goalkeeper. */
void Player::goalieMainLoop( )
{
  brokenGoalieVersion1();
//  deMeer5_goalie();
}    

/*! This is the main decision loop for a defender. */
void Player::defenderMainLoop( )
{
  brokenPlayerVersion1() ;
//    deMeer5();
}

/*! This is the main decision loop for a midfielder. */
void Player::midfielderMainLoop( )
{
  brokenPlayerVersion1() ;
//    deMeer5();
}

/*! This is the main decision loop for an agent. */
void Player::attackerMainLoop( )
{
  brokenPlayerVersion1();
//    deMeer5();
}

void Player::updateAllTimers()
{
  for (int i = 0; i < (int)levels.size(); ++i)
  {
    levels[i]->updateTimer();
  }
}

/*! This method is a first complete team made by David Brokenshire. It is
    based very closely on the Robocup 2001 strategy described in the masters
    thesis by Jelle Kok and the other guy.  This is meant to be a good starting
    point from which we can do some actual science.
*/
void Player::brokenPlayerVersion1()
{
  generateAction();
}

/*! This method is a initial goaltender based off of the description of the
    UVA2001.  Still need to greatly improve method of clearing ball to
    defenders, it currently spends more time going to the other team.
*/
void Player::brokenGoalieVersion1()
{
  SoccerCommand soc;
  VecPosition   posAgent = WM->getAgentGlobalPosition();
  AngDeg        angBody  = WM->getAgentGlobalBodyAngle();
  int i;  //for indexing

  int tIndex = 10;
  // the list of free teammates
  ObjectT     leftBackTeammate;
  ObjectT     leftTeammate;
  ObjectT     leftForwardTeammate;
  ObjectT     forwardTeammate;
  ObjectT     rightForwardTeammate;
  ObjectT     rightTeammate;
  ObjectT     rightBackTeammate;
  ObjectT     backTeammate;
  ObjectT     bestTeammateInArea6;
  ObjectT     bestTeammateInArea6a;

  // teammetes for ordering for each region
  ObjectT     tMates[tIndex];
  for( i=0; i<tIndex; i++)
    tMates[i] = OBJECT_ILLEGAL;

  // define the top and bottom position of a rectangle in which keeper moves
  //static const VecPosition posLeftTop( -PITCH_LENGTH/2.0 +
    //           0.7*PENALTY_AREA_LENGTH, -PENALTY_AREA_WIDTH/4.0 );
  //static const VecPosition posRightTop( -PITCH_LENGTH/2.0 +
    //           0.7*PENALTY_AREA_LENGTH, +PENALTY_AREA_WIDTH/4.0 );
  static const VecPosition posLeftTop( -PITCH_LENGTH/2.0 +
               PENALTY_AREA_LENGTH-1.0, -PENALTY_AREA_WIDTH/2.0+1.0 );
  static const VecPosition posRightTop( -PITCH_LENGTH/2.0 +
               PENALTY_AREA_LENGTH-1.0, +PENALTY_AREA_WIDTH/2.0-1.0 );

  // define the borders of this rectangle using the two points.
  static Line  lineFront = Line::makeLineFromTwoPoints(posLeftTop, posRightTop);
  static Line  lineLeft  = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posLeftTop.getY()), posLeftTop );
  static Line  lineRight = Line::makeLineFromTwoPoints(
                         VecPosition( -50.0, posRightTop.getY()),posRightTop );
  //ObjectT mostFreeForwardTeammateL = getMostFreeForwardTeammateP(posLeftTop);
  //ObjectT mostFreeForwardTeammateR = getMostFreeForwardTeammateP(posRightTop);

  if( WM->isBeforeKickOff( ) )
  {
    if( formations->getFormation() != FT_INITIAL || // not in kick_off formation
        posAgent.getDistanceTo( WM->getStrategicPosition() ) > 2.0 )
    {
      formations->setFormation( FT_INITIAL );       // go to kick_off formation
      ACT->putCommandInQueue( teleportToPos( WM->getStrategicPosition() ) );
    }
    else                                            // else turn to center
    {
      ACT->putCommandInQueue( turnBodyToPoint( VecPosition( 0, 0 ), 0 ) );
      ACT->putCommandInQueue( alignNeckWithBody( ) );
    }
    return;
  }

  if( WM->isFreeKickUs() == true || WM->isGoalKickUs() == true )
  {
    if( WM->isBallKickable() )
    {
      if ( WM->getTimeSinceLastCatch() < 25 && WM->isFreeKickUs() )
      {
         ACT->putCommandInQueue( turnBodyToPoint( VecPosition( 0, 0 ), 0 ) );
         //ACT->putCommandInQueue( alignNeckWithBody( ) );
         if (WM->isLastMessageSee())
            ACT->putCommandInQueue( turnNeckDiffSide( turnBodyToPoint( VecPosition( 0, 0 ), 0 )  ) );
      }
      else if( WM->getTimeSinceLastCatch() == 45 && WM->isFreeKickUs() )
      {
        // move to position with lesser opponents, no
        // move to position where most free teammate is , but make it later
       /* if( WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                                          Circle(posRightTop, 15.0 )) <
            WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                                           Circle(posLeftTop,  15.0 )) )
          soc.makeCommand( CMD_MOVE, posRightTop.getX(),posRightTop.getY(),0.0);
        else
          soc.makeCommand( CMD_MOVE, posLeftTop.getX(), posLeftTop.getY(), 0.0); */

        if( WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                                          Circle(VecPosition(-27.0, PITCH_WIDTH/2.0), 15.0 )) <
            WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                                           Circle(VecPosition(-27.0, -PITCH_WIDTH/2.0),  15.0 )) )
          soc.makeCommand( CMD_MOVE, posRightTop.getX(),posRightTop.getY(),0.0);
        else
          soc.makeCommand( CMD_MOVE, posLeftTop.getX(), posLeftTop.getY(), 0.0);

        ACT->putCommandInQueue( soc );
      }


      else if(WM->getTimeSinceLastCatch() > 50 )
      {
         posAgent=WM->getAgentGlobalPosition();
         mapFreeTeammates(&leftBackTeammate, &leftTeammate, &leftForwardTeammate, &forwardTeammate,
                   &rightForwardTeammate, &rightTeammate, &rightBackTeammate,
                   &backTeammate, &bestTeammateInArea6, &bestTeammateInArea6a,
                   posAgent);
         tMates[0] = forwardTeammate;
         if( posAgent.getY() < 0)
         {
          // tMates[1] = leftForwardTeammate;
          // tMates[2] = leftTeammate;
          tMates[1] = rightForwardTeammate;
          tMates[2] = rightTeammate;
         }
         else
         {
          // tMates[1] = rightForwardTeammate;
          // tMates[2] = rightTeammate;
          tMates[1] = leftForwardTeammate;
          tMates[2] = leftTeammate;
         }

         for(i=0; i<3; i++)
         {
            //cout <<"freedom " <<WM->calculateFreedom(tMates[i])<<" n "<<i<<endl;
            if (  tMates[i]!=OBJECT_ILLEGAL&& WM->isVeryFree(tMates[i] )&&
                  posAgent.getDistanceTo(WM->getGlobalPosition(tMates[i])) < 6.0)

            {
              //cout << " goalie is passing teammate "<<i<<endl;
              if (posAgent.getDistanceTo(tMates[i]) < 6.0) // magic number, some experiments required
                soc = directPass(tMates[i], PASS_SLOW);
              else if(posAgent.getDistanceTo(tMates[i]) < 12.0)
                soc = directPass(tMates[i], PASS_NORMAL);
              else
                soc = directPass(tMates[i], PASS_FAST);

              break;
            }
            else
            {
              soc = SoccerCommand (CMD_ILLEGAL);
            }
         }

         if( soc.isIllegal())
         {
           //cout << " goalie is making clear ball "<<endl;
           soc = kickTo(WM->getPointBetweenTwoClosestOpponentsInCone(posAgent,
                      VecPosition(posAgent.getX()+20.0, PITCH_WIDTH*0.5*sign(posAgent.getY())),
                      VecPosition(0.0,PITCH_WIDTH*0.25*sign(posAgent.getY())) ),
                      //VecPosition(0.0, 0.0)),
                      SS->getBallSpeedMax()) ;

//          Vadim begin
            // the SayBall message with these parameters will only be sent
            // when the player kicks the ball and has seen it
            WM->iSayBallMsgCode     = WM->MSG_I_DRIBBLE;
            WM->iNrCyclesToComplete = 0;                  // this is a DUMMY value
            WM->iPlayerInvolved     = WM->getPlayerNumber() ;
//          Vadim end


         }

        ACT->putCommandInQueue( soc );
      }
      else if(WM->getTimeSinceLastCatch() > 45 )
      {
        posAgent=WM->getAgentGlobalPosition();
        VecPosition posSide( 0.0, posAgent.getY() );
        if( fabs( (posSide - posAgent).getDirection() - angBody) > 10 )
        {
          soc = turnBodyToPoint( posSide );
          ACT->putCommandInQueue( soc );
        }

         if (WM->isLastMessageSee())
            ACT->putCommandInQueue( turnNeckDiffSide( soc  ) );
        //ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if( WM->isGoalKickUs()  )
    {
      if (!WM->isConfidenceGood(OBJECT_BALL))
      {
        ACT->putCommandInQueue( soc = searchBall() );
      }
      else
      {
        ACT->putCommandInQueue( soc = intercept( true, INTERCEPTION_AS_SECOND ) );
        ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
  }
  else
    generateGoalieAction();

  return;
}

/*! This method enacts the goalie behaviour basically as described for UVA2001
    Necessary numerical parameters have been made up, and should be learned
    using some form of learning algorithm in the future.
*/
SoccerCommand Player::UVA2001Goalie()
{
	int fastestTimeToBall = -1;
	int fastestOppTimeToBall = -1;
//if confidence associated with current ball position is too low then
	if (!WM->isConfidenceGood(OBJECT_BALL))
		//return searchBall()
		return searchBall();
	//else if ball is catchable then
	else if (WM->isBallCatchable())
		//return catchBall()
		return catchBall();
	//else if fastest player to ball is clearly me and interception point lies inside penalty area then
	else
	{
		ObjectT fastestPlayer = WM->getFastestInSetTo(OBJECT_SET_PLAYERS, OBJECT_BALL, &fastestTimeToBall);
		WM->getFastestInSetTo(OBJECT_SET_OPPONENTS, OBJECT_BALL, &fastestOppTimeToBall);

		if (fastestPlayer == WM->getAgentObjectType() && WM->isInOwnPenaltyArea(WM->predictPosAfterNrCycles(OBJECT_BALL, fastestTimeToBall)) && fastestTimeToBall - fastestOppTimeToBall < 3) //This magic number, the buffer to make sure the goalie is "clearly" the fastest, should be determined empirically or by learning and placed somewhere, probably in player settings
			//return intercept()
			return intercept(true, INTERCEPTION_AS_SECOND );
		//else if ball is heading towards goal and will enter goal in < 20 cycles then
		else if (WM->isBallHeadingToGoal())
		{
//      return interceptScoringAttempt();

      //Make goalie line
      Line lineGoal = Line::makeLineFromPositionAndAngle(VecPosition( - PITCH_LENGTH/2.0 + PS->getDefendGoalLineDistance(), 0 ), 90 );
      //make ball trajectory
      Line ballTrajectory = Line::makeLineFromPositionAndAngle( WM->getBallPos(), WM->getBallDirection());
      //find intersection (get z)
			// ~z = desired guard point, t + t = agent's global body angle in cycle t
			//~z = intersection(ball trajectory, goalie line)
      VecPosition z = lineGoal.getIntersection(ballTrajectory);

      //Find the correct angle to make goalie look in direction of ball when standing on interception point.
      AngDeg desiredGlobalBodyDir = (WM->getBallPos() - z).getDirection(); //I believe this is the correct global body angle
      //return moveToPosAlongLine(~z, t + t, GoalieDevDist, 1, GoalieTurnThr, GoalieCorrAng)
      return moveToPosAlongLine(z, desiredGlobalBodyDir, 3.0, 1, 7, 12); //These magic numbers should be put into player settings

    }
    else
    {
      return defendGoalLine(PS->getDefendGoalLineDistance());
/*
      //Ible: My addition to deal with stamina problems: ***NEEDS LOTS MORE WORK***
      if (WM->getRelativeDistance(OBJECT_BALL) > 35)
      {
        //If we are more than a few degrees of 90, turn to 90
        if (fabs(fabs(WM->getAgentGlobalBodyAngle()) - 90) > 5)
        { int multiplier = 1;
          if (WM->getAgentGlobalBodyAngle() < 0) multiplier = -1;
            return turnBodyToPoint(VecPosition(-PITCH_LENGTH/2.0 + PS->getDefendGoalLineDistance(), multiplier*SS->getGoalWidth()/2.0));
        }
        //otherwise do nothing
      }


      if (WM->getAgentTiredNess() == TIREDNESS_GOOD)
        //return defendGoalLine(DefendGoalLineDist)
        return defendGoalLine(PS->getDefendGoalLineDistance());
      else
      {
        if (WM->getRelativeDistance(OBJECT_BALL) > 25)
          return this->turnBodyToObject(OBJECT_BALL);
        else
          return defendGoalLine(PS->getDefendGoalLineDistance());
      }
*/
    }
  }
//end if
}


/*!This method determines whether a player should say something. 
   \return bool indicating whether the agent should say a message */
bool Player::shallISaySomething( )
{
  bool        bReturn;

  bReturn  = ((WM->getCurrentTime() - m_timeLastSay) >= SS->getHearDecay());
  bReturn  &= amIAgentToSaySomething();
  bReturn  &= (WM->getCurrentCycle() > 0 );

  return bReturn;
}

/*! This method returns a boolean indicating whether I should communicate my
    world model to the other agents. 
    \return boolean indicating whether I should communicate my world model. */
bool Player::amIAgentToSaySomething()
{
  // if got good information of the ball, communicate. 
  if( WM->getTimeChangeInformation(OBJECT_BALL) == WM->getCurrentTime() ||
      WM->getRelativeDistance( OBJECT_BALL ) < SS->getVisibleDistance()    ) 
    return true;
    
  return false;
}


/*!This method does nothing except updating the worldmodel after each arrival of
   a see or sense_body message. Using this test program it is possible to
   control the commands of the agent using the standard input and see the
   resulting behavior of the agent. */
void Player::test_only_update()
{
  while( bContLoop )
  {
    WM->updateAll();
    if(  WM->waitForNewInformation() == false )
    {
        printf("Server dead; exiting" );
        bContLoop =  false;
    }
  }
}

/*!This method can be called in a separate thread (see pthread_create) since
   it returns a void pointer. It is assumed that this function receives a
   BasicPlayer class as argument. The only thing this function does
   is starting the method handleStdin() from the corresponding BasicPlayer
   class that listens to user input from the keyboard. This function is
   necessary since a method from a class cannot be an argument of
   pthread_create.
   \param v void pointer to a BasicPlayer class.
   \return should never return since function handleStdin has infinite loop*/
void* stdin_callback( void * v )
{
  Log.log( 1, "Starting to listen for user input" );
  Player* p = (Player*)v;
  p->handleStdin();
  return NULL;
}

/*!This method listens for input from the keyboard and when it receives this
   input it converts this input to the associated action. See
   showStringCommands for the possible options. This method is used together
   with the SenseHandler class that sends an alarm to indicate that a new
   command can be sent. This conflicts with the method in this method that
   listens for the user input (fgets) on Linux systems (on Solaris this isn't a
   problem). The only known method is to use the flag SA_RESTART with this
   alarm function, but that does not seem to work under Linux. If each time
   the alarm is sent, this gets function unblocks, it will cause major
   performance problems. This function should not be called when a whole match
   is played! */
void Player::handleStdin( )
{
  char buf[MAX_MSG];

  while( bContLoop )
  {
    fgets( buf, MAX_MSG, stdin ); // does unblock with signal !!!!!
    printf( "after fgets: %s\n", buf );
    executeStringCommand( buf );
  }
}

/*!This method prints the possible commands that can be entered by the user.
   The whole name can be entered to perform the corresponding command, but
   normally only the first character is sufficient. This is indicated by
   putting brackets around the part of the command that is not needed.
   \param out output stream to which the possible commands are printed */
void Player::showStringCommands( ostream& out )
{
  out << "Basic commands:"                                << endl <<
         " a(ctions)"                                     << endl <<
         " c(atch) direction"                             << endl <<
         " cs(lientsettings"                              << endl <<
         " d(ash) power [ times ]"                        << endl <<
         " de(bug) nr_cycles"                             << endl <<
         " g(oto) x y"                                    << endl <<
         " h(elp)"                                        << endl <<
         " i(ntercept) x y"                               << endl <<
         " k(ick) power angle"                            << endl <<
         " ka x y endspeed "                              << endl <<
         " m(ove) x y"                                    << endl <<
         " n(eck) angle"                                  << endl <<
         " o(pponents in cone) width dist"                << endl <<
         " p(redict cycles to) x y"                       << endl <<
         " q(uit)"                                        << endl <<
         " s(ay) message"                                 << endl <<
         " ss(erversettings)"                             << endl <<
         " t(urn) angle"                                  << endl <<
         " v(iewmode) narrow | normal | wide low | high"  << endl <<
         " w(orldmodel)"                                  << endl;
}

/*!This method executes the command that is entered by the user. For the
   possible command look at the method showStringCommands.
   \param str string that is entered by the user
   \return true when command could be executed, false otherwise */
bool Player::executeStringCommand( char *str)
{
  SoccerCommand socCommand;
  int           i;
  double        x, y;

  switch( str[0] )
  {
    case 'a':                                 // actions
      WM->showQueuedCommands();
      break;
    case 'c':                                 // catch dir or cs
      if( strlen(str) > 1 && str[1] == 's' )
      {
        PS->show( cout, ":" );
        break;
      }
      socCommand.makeCommand( CMD_CATCH, Parse::parseFirstInt( &str ) );
      break;
    case 'd':                                 // dash
      socCommand.commandType = CMD_DASH;
      socCommand.dPower      = Parse::parseFirstDouble( &str );
      socCommand.iTimes      = Parse::parseFirstInt   ( &str );
      if( socCommand.iTimes == 0 ) socCommand.iTimes = 1;
      break;
    case 'h':                                // help
      showStringCommands( cout );
      return true;
    case 'k':                                // kick or ka (kick advanced)
      socCommand.commandType = CMD_KICK;
      if( str[1] == 'a' ) // advanced kick
      {
        double x = Parse::parseFirstDouble( &str );
        double y = Parse::parseFirstDouble( &str );
        double e = Parse::parseFirstDouble( &str );
        socCommand = kickTo( VecPosition( x, y), e );
      }
      else
      {
        socCommand.dPower = Parse::parseFirstDouble( &str );
        socCommand.dAngle = Parse::parseFirstDouble( &str );
      }
      break;
    case 'm':                               // move
      socCommand.commandType = CMD_MOVE;
      socCommand.dX          = Parse::parseFirstDouble( &str );
      socCommand.dY          = Parse::parseFirstDouble( &str );
      socCommand.dAngle      = Parse::parseFirstDouble( &str );
      break;
    case 'n':                              // turn_neck
      socCommand.commandType = CMD_TURNNECK;
      socCommand.dAngle      = Parse::parseFirstDouble( &str );
      break;
    case 'o':                              // count nr opp in cone
      x = Parse::parseFirstDouble( &str );
      y = Parse::parseFirstDouble( &str );
      i = WM->getNrInSetInCone( OBJECT_SET_OPPONENTS, x, WM->getAgentGlobalPosition(),
              WM->getAgentGlobalPosition()+VecPosition( y,
          WM->getAgentGlobalNeckAngle(), POLAR ) );
      printf( "%d opponents\n", i );
      return true;
    case 'p':                              // predict cycles to point
      x = Parse::parseFirstDouble( &str );
      y = Parse::parseFirstDouble( &str );
      i = WM->predictNrCyclesToPoint( WM->getAgentObjectType(),
                     VecPosition( x, y ), PS->getPlayerWhenToTurnAngle() );
      printf( "%d cycles\n", i );
      return true;
    case 'q':                             // quit
      bContLoop = false;
      return true;
    case 's':                             // ss (serversettings) or say
      if( strlen(str) > 1 && str[1] == 's' )
      {
        SS->show( cout, ":" );
        break;
      }
      socCommand.commandType = CMD_SAY;
      Parse::gotoFirstOccurenceOf( ' ', &str );
      Parse::gotoFirstNonSpace( &str );
      strcpy( socCommand.str, str);
      break;
    case 't':                             // turn
      socCommand.commandType = CMD_TURN;
      socCommand.dAngle      = Parse::parseFirstDouble( &str );
      break;
    case 'v':                             // change_view
      socCommand.commandType = CMD_CHANGEVIEW;
      Parse::gotoFirstOccurenceOf(' ', &str );
      Parse::gotoFirstNonSpace( &str );
      socCommand.va          = SoccerTypes::getViewAngleFromStr( str );
      Parse::gotoFirstOccurenceOf(' ', &str );
      Parse::gotoFirstNonSpace( &str );
      socCommand.vq          = SoccerTypes::getViewQualityFromStr( str );
      break;
    case 'w':                            // worldmodel
      WM->show();
      return true;
    default:                             // default: send entered string
      ACT->sendMessage( str );
      return true;
  }
  if( socCommand.commandType != CMD_ILLEGAL ) // when socCommand is set
    ACT->putCommandInQueue( socCommand );     // send it.

  return true;
}


/*! This method picks an action mode based on some simple conditions, the action
    mode is then used to call functions to select a specific action.

  \return enum ActionT indicating which type of action is appropriate
*/
ActionT Player::determineActionMode()
{
  //Before kick off
  if (WM->isBeforeKickOff())
    return ACT_BEFORE_KICKOFF;
  else if (!WM->isConfidenceGood(OBJECT_BALL))
    return ACT_SEARCH_BALL;
  else if (WM->isDeadBallThem())
    return ACT_MOVE_TO_DEAD_BALL_POSITION;
  else if (WM->isBallKickable())
    return ACT_KICK_BALL;
  else if (WM->getFastestInSetTo(OBJECT_SET_TEAMMATES, OBJECT_BALL)
           == WM->getAgentObjectType() && intercept(false, INTERCEPTION_AS_SECOND ).commandType!=CMD_ILLEGAL)
    return ACT_INTERCEPT;
  else if (WM->isDeadBallUs())
    return ACT_MOVE_TO_DEAD_BALL_POSITION;
  else if (shouldIMark())
    return ACT_MARK;
  else if (WM->getAgentGlobalPosition().getDistanceTo(WM->getStrategicPosition()) > 1)
    return ACT_GOTO_STRATEGIC_POSITION;
  else
    return ACT_WATCH_BALL;
}

/*! This method for checking positioning

  \return enum ActionT indicating which type of action is appropriate
*/
ActionT Player::determineActionMode1()
{
/*
  if ( WM->multiCycleAction == ACT_INTERCEPT_PASS )
    cout << " #" << WM->getPlayerNumber()
         << " multiCycleAction = ACT_INTERCEPT_PASS "
         << " cycle = " << WM->getCurrentCycle()
         << endl;
*/

  //Before kick off
  InterceptionT iType;
  
  if (WM->isBeforeKickOff())
  {
    if (WM->getAgentViewAngle() !=VA_WIDE)
    {
      ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_WIDE, VQ_HIGH ) );
    }  
    return ACT_BEFORE_KICKOFF;
  }
  else if (!WM->isConfidenceGood(OBJECT_BALL))
  {
    if (WM->getAgentViewAngle() !=VA_NORMAL)
    {
      ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_NORMAL, VQ_HIGH ) );
    }  
    //cout<<"search "<<WM->getPlayerNumber()<<"cycle "<<WM->getCurrentCycle()<<endl;
    return ACT_SEARCH_BALL;
  }

  else if (WM->isDeadBallThem())
  {
    if (WM->getAgentViewAngle() !=VA_WIDE)
    {
      ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_WIDE, VQ_HIGH ) );
    } 
    return ACT_MOVE_TO_DEAD_BALL_POSITION;
  }
  
  else if(WM->getTimeSinceLastCatch() < PS->getCyclesCatchWait() ||
        WM->isGoalKickUs() )
  {
        return ACT_MOVE_TO_DEAD_BALL_POSITION;
  }
   
  else if (WM->isBallKickable())
  {
    //cout<<"kick "<<WM->getPlayerNumber()<<"cycle "<<WM->getCurrentCycle()<<endl;
    if (WM->getAgentViewAngle() !=VA_NORMAL)
    {
      ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_NORMAL, VQ_HIGH ) );
    }
    return ACT_KICK_BALL;

  }
/*
  else if ( WM->multiCycleAction == ACT_INTERCEPT_PASS  && false )     // DISABLED
  {

    cout << "&&&& #" << WM->getPlayerNumber()
         << ": Intercepting in spite of all doubts." << endl;

    return ACT_INTERCEPT_PASS;
  }
*/
  else if ( shouldIIntercept( &iType) )
  
  {
    //cout<<"intercept "<<WM->getPlayerNumber()<<"cycle "<<WM->getCurrentCycle()<<endl;
    if (WM->getAgentViewAngle() !=VA_NORMAL)
    {
      ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_NORMAL, VQ_HIGH ) );
    }
    return ACT_INTERCEPT;
  }

  else if ( WM->isDeadBallUs() )

  {
    if ( WM->getAgentViewAngle() !=VA_WIDE )
    {
      ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_WIDE, VQ_HIGH ) );
    } 
    return ACT_MOVE_TO_DEAD_BALL_POSITION;
  }

  else if ( shouldIMark1(VecPosition( 0.0, 0.0 ) ) )

  {
    //cout<<"mark "<<WM->getPlayerNumber()<<"cycle "<<WM->getCurrentCycle()<<endl;
    return ACT_MARK;
   }
  /*else if (WM->getAgentGlobalPosition().getDistanceTo(WM->getStrategicPosition()) > 8)
  //else if (WM->getAgentGlobalPosition().getDistanceTo(
  //getAnticipPosition(WM->getStrategicPosition(), 8.0)) > 1.5)
  {
    //cout<<"strtpos "<<WM->getPlayerNumber()<<"cycle "<<WM->getCurrentCycle()<<endl;
    return ACT_GOTO_STRATEGIC_POSITION;
  }

  else if (WM->getTeamPossesion() != POSSESION_US && shouldIMark() )
  {
    //cout<<"mark "<<WM->getPlayerNumber()<<"cycle "<<WM->getCurrentCycle()<<endl;
    return ACT_MARK;
   }
  else
    return ACT_WATCH_BALL; */
  else
  {
    if (WM->getAgentViewAngle() !=VA_NORMAL)
    {
      ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_NORMAL, VQ_HIGH ) );
    }
    return ACT_GOTO_STRATEGIC_POSITION;
  }

} // determineActionMode1



/*! This method maps from action mode to action command by either by calling a
    skill directly, or by calling an additional decision procedure to decide
    what skill should be called.  Currently ACT_KICK_BALL is the only mode
    which uses a decision sub-function.  This method is also responsible for
    sending the command it generates to the action handler
*/
bool Player::generateAction()
{
  //debug for target area
 //   if (WM->getCurrentCycle()%50==1)
 // {
  // cout<<WM->getCurrentCycle()<<"P # "<<WM->getPlayerNumber()<<" ";
  // cout<<"APos "<<WM->getAgentGlobalPosition();
  // cout<<"BPos "<<WM->getBallPos()<<" "<<WM->getConfidence(OBJECT_BALL);
  // WM->setTargetArea();
  // WM->printDebug();
  //}//end of debug

  ActionT mode = determineActionMode();
  SoccerCommand action;
  if (mode == ACT_BEFORE_KICKOFF)
  {
    setUpForKickOff();
    return true;
  }
  else if (mode == ACT_MOVE_TO_DEAD_BALL_POSITION)
  {
    formations->setFormation( FT_433_OFFENSIVE );
    if (WM->getAgentGlobalPosition().getDistanceTo(WM->getStrategicPosition()) >= 5)
      action = moveToPos(WM->getStrategicPosition(), 20, 2); //Magic numbers
  }
  else if (mode == ACT_SEARCH_BALL)
  {
    formations->setFormation( FT_433_OFFENSIVE );
    action = searchBall();
  }
  else if (mode == ACT_KICK_BALL)
  {
    formations->setFormation( FT_433_OFFENSIVE );
    action = determineActionWithBall();
  }
  else if (mode == ACT_INTERCEPT)
  {
    formations->setFormation( FT_433_OFFENSIVE );
    action = intercept(false, INTERCEPTION_AS_SECOND );
  }
  else if (mode == ACT_MARK)
  {
    formations->setFormation( FT_433_OFFENSIVE );
    action = mark(opponentToMark,PS->getMarkDistance(), MARK_BALL);
  }
  else if (mode == ACT_GOTO_STRATEGIC_POSITION)
  {
    formations->setFormation( FT_433_OFFENSIVE );
    if (WM->getAgentGlobalPosition().getDistanceTo(WM->getStrategicPosition()) >= 5)
      action = moveToPos(WM->getStrategicPosition(), 20, 2); //Magic numbers
  }
  else if (mode == ACT_WATCH_BALL)
  {
    formations->setFormation( FT_433_OFFENSIVE );
    action = turnBodyToObject(OBJECT_BALL);
  }
  else
  {
    Log.logWithTime(777, "Error condition in generateAction! Unknown ACT_\n");
  }

  action = adjustDashPowerForStamina(action, mode);
  ACT->putCommandInQueue(action);
  if (WM->isLastMessageSee())
    ACT->putCommandInQueue(turnNeckDiffSide(action));
  else
    ACT->putCommandInQueue(turnNeckToObject(OBJECT_BALL,action));

  // added for debugging first version with levels. Always gives control to highest level
  // so highest level is executed every cycle
  //currentLevel = levels.size()-1;

  return true;
}


/*! This method is used for level 1. Action is selected based on decision of
    level 2
    This method is also responsible for
    sending the command it generates to the action handler
*/
LevelStateT Player::generateAction1()
{

  if ( WM->multiCycleAction == ACT_INTERCEPT_PASS )
    cout << " #" << WM->getPlayerNumber()
         << " multiCycleAction = ACT_INTERCEPT_PASS "
         << "  step = " << multiCycleDribbleStep
         << "  cycle = " << WM->getCurrentCycle()
         << endl;

  LevelStateT lst = ACTIONS_SELECTED_FINISHED;  // indicates the result of the action
                                                // for all one-cycle actions always
                                                // ACTIONS_SELECTED_FINISHED
  SoccerCommand action = SoccerCommand(CMD_ILLEGAL);
  double dist;
  VecPosition pointPass;
  VecPosition posStrat;
  int Cycle = 0;


  if (WM->getBallPossesion()!=POSSESION_US && !WM->isInOffenseArea(OBJECT_BALL))
      formations->setFormation(FT_433_DEFENSIVE);
  else
      formations->setFormation( FT_433_OFFENSIVE );

  // start perform actions

  switch (currentAction)
  {
    case ACT_BEFORE_KICKOFF:

     setUpForKickOff();
     lst = ACTIONS_SELECTED_FINISHED;
     if (WM->isLastMessageSee())
      ACT->putCommandInQueue(turnNeckDiffSide(action));
      //action = turnNeckDiffSide(action);
     return lst;
    break;

    case ACT_MOVE_TO_DEAD_BALL_POSITION:

      lst = ACTIONS_SELECTED_FINISHED;
      if (WM->isGoalKickUs() || ( WM->isFreeKickUs()&&
                               WM->getTimeSinceLastCatch()< 29))
      {

        formations->setFormation(FT_433_DEFENSIVE);
      }

      if( WM->isBallKickable() )
      {
        posStrat = WM->getPosForDeadBall();
        dist = 0.5;
      }
      else
      {
        posStrat = WM->getStrategicPosition();
        dist = 5.0;
      }

      if (WM->getAgentGlobalPosition().getDistanceTo(posStrat) > dist &&
       //   WM->getAgentTiredNess() != TIREDNESS_VERY_BAD &&
       //   WM->getAgentTiredNess() != TIREDNESS_BAD)
      ( WM->getAgentStamina().getStamina() > 3000 && WM->isDeadBallUs() ||
        WM->getAgentStamina().getStamina() > 2000 && WM->isDeadBallThem()) )
      {
       action = moveToPos(posStrat, 20, 2); //Magic numbers
       lst = ACTIONS_SELECTED_FINISHED;
      }
      
      /*
      else
      {
        posStrat = WM->getAdjustedStratPosition(WM->getPlayerNumber(), WM->getAgentGlobalPosition());
        if (WM->getAgentGlobalPosition().getDistanceTo(posStrat) > 1.5 )
        {
          action = moveToPos(posStrat, 20, 2); //Magic numbers
          lst = ACTIONS_SELECTED_FINISHED;
        }*/
        else
        {
          action = turnBodyToObject(OBJECT_BALL);
          lst = ACTIONS_SELECTED_FINISHED;
       // }
      }
    break;

    case ACT_SEARCH_BALL:

     action = searchBall();
     ACT->putCommandInQueue(action);
     lst = ACTIONS_SELECTED_FINISHED;
     return lst;
    break;


  // actions with Ball
    case ACT_TURN_WITH_BALL:

      lst = ACTIONS_SELECTED_FINISHED;
      action = turnWithBallTo(dribbleDirection,PS->getTurnWithBallAngThr(),PS->getTurnWithBallFreezeThr());
    break;

    case ACT_DRIBBLE:

     return dribbleSimpleMultiCycle(dribbleDirection, dribbleType, pointToPass);
     lst = ACTIONS_SELECTED_FINISHED;
     // this method is implemented but requires some extra work
    break;

    case ACT_TRICK:

     return trickOppMultyCycle( teammateToPass );
     lst = ACTIONS_SELECTED_FINISHED;

    break;

    case ACT_HIDE_BALL:

     return hideBallMultiCycle( teammateToPass );
     lst = ACTIONS_SELECTED_FINISHED;
     
    break;
    
    case ACT_OUTPLAY:

     return outplaySimpleMultiCycle( pointToPass );
     lst = ACTIONS_SELECTED_FINISHED;
     // this method is implemented as a simple kick between two closest opponents
    break;


    case ACT_DIRECT_PASS:

      if ( WM->multiCycleAction == ACT_PASS_ON_REQUEST )
      {
/*
        cout << "*** #" << WM->getPlayerNumber()
        << " passing on request by #"
        << WM->iPlayerInvolvedHeard
        << "  time = " << WM->getCurrentCycle()
        << endl;
*/
        PassT passType;
        if ( WM->vpSayBallPointHeard.getDistanceTo( WM->getBallPos() ) > 12.0 )
          passType = PASS_FAST;
        else
          passType = PASS_NORMAL;  
        return directPassMultiCycle( WM->vpSayBallPointHeard,
                                     WM->iPlayerInvolvedHeard,
                                     passType );             
      }
      else
      {
        pointPass = WM->getGlobalPosition(teammateToPass);
        if ( pointPass.getDistanceTo(WM->getBallPos()) > 12.0 )  // magic number
          //action = directPassMultiCycle(teammateToPass,PASS_FAST);
          return  directPassMultiCycle(teammateToPass, PASS_FAST);
        else
          //action = directPassMultiCycle(teammateToPass,PASS_NORMAL);
          return  directPassMultiCycle(teammateToPass, PASS_NORMAL);
      }
      lst = ACTIONS_SELECTED_FINISHED;
    break;

    case ACT_LEADING_PASS:

      return leadingPassMultiCycle(teammateToPass);
       //action = leadingPass(teammateToPass,3.0); //magic number
       //lst = ACTIONS_SELECTED_FINISHED;
    break;

    case ACT_KICK_TO:

     //   cout<<"this is kick "<<WM->getPlayerNumber()<<endl;
     if (WM->isKickOffUs())
     {
       action  = kickTo(pointToPass, PS->getPassEndSpeed()/2.0); // magic number
     }
     else
     {
      action  = kickTo(pointToPass, SS->getBallSpeedMax());
     }
     lst = ACTIONS_SELECTED_FINISHED;
    break;

    case ACT_INTERCEPT:

     lst = ACTIONS_SELECTED_FINISHED;
     return interceptSimpleMultiCycle( false );

    break;

/*    
    case ACT_INTERCEPT_PASS:

    cout << " #" << WM->getPlayerNumber()
         << " ACT_INTERCEPT_PASS "
         << " cycle = " << WM->getCurrentCycle()
         << endl;

     lst = ACTIONS_SELECTED_FINISHED;
     return inteceptPassMultiCycle( WM->vpSayBallPoint );

    break;
*/     

    case ACT_GOTO_STRATEGIC_POSITION:

     if ( shouldIRequestForPassing () )
     { // do nothing; everything is done inside this method anyway
     }

    /* if (WM->isGoalKickUs())
     {
       //cout<<"set form in pos"<<endl;
      formations->setFormation( FT_ATTACKING );
     } */

    posStrat = WM->getStrategicPosition();

    Cycle = WM->getCurrentCycle();

    if (WM->getAgentGlobalPosition().getDistanceTo( posStrat) > 5.0 )
    {
       action = moveToPos(posStrat, 20, 2); //Magic numbers

    }
  
    else if ( (WM->isInDefenseArea(OBJECT_BALL) || WM->getPlayerNumber() > 8 )&&
             WM->getAgentGlobalPosition().getDistanceTo( posStrat) > 2.5)
    {
      action = moveToPos(posStrat, 20, 2); //Magic numbers
    }
    else if ( (formations->getPlayerType() == PT_DEFENDER_WING ||
             formations->getPlayerType() == PT_DEFENDER_SWEEPER) &&
             WM->getAgentGlobalPosition().getX() > posStrat.getX() +1.5)
    {         
      action = moveToPos(VecPosition(posStrat.getX(),WM->getAgentGlobalPosition().getY()) , 20, 2); //Magic numbers
    }
    else if( WM->getAgentGlobalPosition().getX() > WM->getOffsideX() )
    {
      action = moveToPos(VecPosition(WM->getOffsideX()-0.5,WM->getAgentGlobalPosition().getY()), 20, 2); //Magic numbers
    }
    else
    {
      /*
        posStrat = WM->getAdjustedStratPosition(WM->getPlayerNumber(), WM->getAgentGlobalPosition());
        if (WM->getAgentGlobalPosition().getDistanceTo(posStrat) > 1.5 )
        {
          action = moveToPos(posStrat, 20, 2); //Magic numbers
          lst = ACTIONS_SELECTED_FINISHED;
        }
        else
        { */
          action = turnBodyToObject(OBJECT_BALL);
          ACT->putCommandInQueue(action);
          ACT->putCommandInQueue(turnNeckDiffSide(action));
          return   ACTIONS_SELECTED_FINISHED;
       // }


    }

    if (Cycle < WM->getCurrentCycle())
      cout<<"looping inside generate action "<<Cycle<<" "<<WM->getCurrentCycle()<<" "<<WM->getPlayerNumber()<<endl;

    lst = ACTIONS_SELECTED_FINISHED;

    break;
    
    case ACT_MARK:
     //cout <<" I am marking"<<endl;
     lst = ACTIONS_SELECTED_FINISHED;
     action = mark(opponentToMark, SS->getPlayerSize()*2.0, MARK_BISECTOR);

    break;

    
    case ACT_WATCH_BALL:

     action = turnBodyToObject(OBJECT_BALL);
     lst = ACTIONS_SELECTED_FINISHED;
    break;


    default:

      Log.logWithTime(777, "Error condition in generateAction! Unknown ACT_\n");

     lst = ACTIONS_UNSELECTED_FAILED;
    break;

  } // end of switch
  
  action = adjustDashPowerForStamina( action, currentAction, false );
 

  if ( action.commandType > 12 || action.commandType < 1 )
    cout<<" here illegal command!!!"<<endl;



  // different variants of change vision, is not finished
  ACT->putCommandInQueue(action);

  if (WM->isLastMessageSee())
  {
    if( currentAction == ACT_MOVE_TO_DEAD_BALL_POSITION ||
        currentAction == ACT_BEFORE_KICKOFF || currentAction == ACT_MARK)
    {
      ACT->putCommandInQueue(turnNeckDiffSide(action));
    }
    else
    {
      ACT->putCommandInQueue(turnNeckToObject(OBJECT_BALL,action));
    }
  }


  return lst;
}



/*! Based on a variety of simple conditions this method decides what mode
    the goalie should act in.  This determines at a high level what kind of
    action the goalie will take.
*/
ActionT Player::determineGoalieActionMode()
{
  Log.logWithTime(777, "Entering determineGoalieActionMode()");
	int fastestTimeToBall = -1;
	int fastestOppTimeToBall = -1;
  int fastestTmateToBall  = -1;
  VecPosition posBall = WM->getBallPos();
  ObjectT fastestTeammate =  WM->getFastestInSetTo(OBJECT_SET_TEAMMATES, OBJECT_BALL, &fastestTmateToBall);
  ObjectT fastestPlayer = WM->getFastestInSetTo(OBJECT_SET_PLAYERS, OBJECT_BALL, &fastestTimeToBall);
  WM->getFastestInSetTo(OBJECT_SET_OPPONENTS, OBJECT_BALL, &fastestOppTimeToBall);
  //Circle c = Circle(WM->getPosOwnGoal(), WM->getBallPos().getDistanceTo(WM->getPosOwnGoal()) );
  //int nrInCircle = WM->getNrInSetInCircle(OBJECT_SET_PLAYERS, c);

  if (!WM->isConfidenceGood(OBJECT_BALL))
    return ACT_SEARCH_BALL;

  else if (WM->isBallCatchable())
    return ACT_CATCH_BALL;
  else if (WM->isBallKickable())
    return ACT_KICK_TO;
  else if (WM->isBallHeadingToGoal())     //Ible: Moved this up from the second last condition because it seemed important.
    return ACT_INTERCEPT_SCORING_ATTEMPT;

   // if the ball inside penalty area we can calculate precisely
 /* else if (nrInCircle <2 &&
          WM->getBallPos().getDistanceTo(VecPosition(-PITCH_LENGTH/2, 0.0)) < 24.0 &&
          WM->getTeamPossesion() == POSSESION_THEM)
  {

      cout << "Using circle interception condition.\n";
        return ACT_INTERCEPT;

  }  */
  else if(WM->isCrossPass())
    return ACT_INTERCEPT_CROSS_PASS;
  // if the goalie is the fastest player - intercept. Serg Apr.17 2003
  else if (fastestPlayer == WM->getAgentObjectType() && fabs(posBall.getY()) < 8.0)
  {
    //cout << "The goalie is the fastest"<<endl;
        return ACT_INTERCEPT;
  }
  else if (fastestPlayer == WM->getAgentObjectType() && fastestOppTimeToBall > fastestTimeToBall-1)
  {
   // cout << "The goalie is the fastest - 1"<<endl;
        return ACT_INTERCEPT;
  }

  else if (fastestTeammate == WM->getAgentObjectType()&&
            WM->getBallPos().getDistanceTo(VecPosition(-PITCH_LENGTH/2, 0.0)) < 7.0)
  {
    //cout << "Using narrowed reitroduced interception condition.\n";
        return ACT_INTERCEPT;
  }

  else if(isItBreakIn(fastestTeammate, fastestPlayer, posBall) )
  {
    //cout<<"Break in"<<endl;
       return ACT_INTERCEPT;
  }

  //Ible: Reintroduced this condition to try to improve play (March 25th '03)
  else if (fastestTeammate == WM->getAgentObjectType()&&
            WM->getBallPos().getDistanceTo(VecPosition(-PITCH_LENGTH/2, 0.0)) < 17.0 &&
            fastestOppTimeToBall < 6 && fastestOppTimeToBall > 1 && fabs(posBall.getY()) < 8.0 )
  {
    //cout << "Using reitroduced interception condition.\n";
        return ACT_INTERCEPT;
  }

           //&& WM->isInOwnPenaltyArea(OBJECT_BALL))
//           &&  fastestTimeToBall - fastestOppTimeToBall < 5)
 //   return ACT_INTERCEPT;
 // else if (fastestPlayer == WM->getAgentObjectType()
 //          && WM->isInOwnPenaltyArea(WM->predictPosAfterNrCycles(OBJECT_BALL,
 //          fastestTimeToBall)) && fastestTimeToBall - fastestOppTimeToBall < 0) //This magic number, the buffer to make sure the goalie is "clearly" the fastest, should be determined empirically or by learning and placed somewhere, probably in player settings
 //   return ACT_INTERCEPT;

  /* intercept outside penalty area
  else if (fastestPlayer == WM->getAgentObjectType()
           && fastestTimeToBall - fastestOppTimeToBall < 0 ) //This magic number, the buffer to make sure the goalie is "clearly" the fastest, should be determined empirically or by learning and placed somewhere, probably in player settings
    return ACT_INTERCEPT;*/

 // else if (fastestTeammate == WM->getAgentObjectType() &&
 //          WM->isInOwnPenaltyArea(WM->predictPosAfterNrCycles(OBJECT_BALL,
 //          fastestTimeToBall)) )
/*<<<<<<< Player.C
    return ACT_INTERCEPT;
    // try to intercept one by one
  else if (fastestPlayer != fastestTeammate && WM->getBallPos().getDistanceTo(WM->getPosOwnGoal()) < 18.0)
    return ACT_INTERCEPT;
  else if (WM->isBallHeadingToGoal())
    return ACT_INTERCEPT_SCORING_ATTEMPT;
=======*/
  //  return ACT_INTERCEPT;
//>>>>>>> 1.7
  else
    return ACT_DEFEND_GOALLINE;

}

/*! This method provides most of the action selection for the goalie.  After
    determining what mode to play in using determineGoalieActionMode the
    method calls the appropriate low level skills to act. This method is
    based closely off the UVATrilearn 2001 (Seattle) team goalie as described
    in the masters thesis.
*/
bool Player::generateGoalieAction()
{
  Log.logWithTime(777, "Entering generateGoalieAction");
  ActionT mode = determineGoalieActionMode();
  SoccerCommand action;
  ObjectT freeTeammate = getMostFreeForwardTeammate();
  if (mode == ACT_SEARCH_BALL)
  {
    action = searchBall();
    //ACT->putCommandInQueue(turnNeckDiffSide(action));
    // cout << "Searching for ball\n";
    Log.logWithTime(777, "Search for ball");
  }
  else if (mode == ACT_KICK_TO)
  {
    if (WM->isVeryFree(freeTeammate))
     //  Vadim begin
    {
      action =  directPass( freeTeammate, PASS_NORMAL);

    }
//  Vadim end
    else if (WM->getBallPos().getY() < 0.0)
      action = kickTo(WM->getPointBetweenTwoClosestOpponentsInCone(
                      WM->getAgentGlobalPosition(),
                      VecPosition(0.0, -PITCH_WIDTH/2),
                      VecPosition(PITCH_LENGTH/2, 0.0)),
                      SS->getBallSpeedMax());
    else
      action = kickTo(WM->getPointBetweenTwoClosestOpponentsInCone(
                      WM->getAgentGlobalPosition(),
                      VecPosition(PITCH_LENGTH/2, 0.0),
                      VecPosition(0.0, PITCH_WIDTH/2)),
                      SS->getBallSpeedMax());
    //ACT->putCommandInQueue(turnNeckToPoint(WM->getBallPos(), action));
    Log.logWithTime(777, "Kick ball");
    //cout << "Kicking ball\n";
  }
  else if (mode == ACT_CATCH_BALL)
  {
    action = catchBall();
    ACT->putCommandInQueue(turnNeckToPoint(WM->getBallPos(), action));
    Log.logWithTime(777, "Catch ball");
    //cout << "Catching ball\n";
  }
  else if (mode == ACT_INTERCEPT_SCORING_ATTEMPT)
  {
    action = goalieInterceptOnLine();
   //ACT->putCommandInQueue(turnNeckToPoint(WM->getBallPos(), action));
   // changed as an attempt to improve intrcept scoring attempt (it seems doesn't work properly)
   ACT->putCommandInQueue(turnNeckToObject(OBJECT_BALL, action));
    Log.logWithTime(777, "Inercept scoring attempt");
    //cout << "Intercepting scoring attempt\n";
  }
  else if (mode == ACT_INTERCEPT_CROSS_PASS)
  {
    action = goalieInterceptCrossPass();
    //action = intercept(false);
    ACT->putCommandInQueue(turnNeckToPoint(WM->getBallPos(), action));
    Log.logWithTime(777, "Inercept cross pass");
    //cout << "Intercepting cross pass"<<endl;
  }
  else if (mode == ACT_INTERCEPT)
  {
    action = intercept(false, INTERCEPTION_AS_SECOND );
    //action = intercept(false);
    ACT->putCommandInQueue(turnNeckToPoint(WM->getBallPos(), action));
    Log.logWithTime(777, "Interception");
   // cout<<WM->getCurrentCycle();
   // cout << " Intercepting\n";
  }
  else if (mode == ACT_DEFEND_GOALLINE)
  {
    if (WM->getBallPos().getX() < -PITCH_LENGTH/2+PS->getDefendGoalLineDistance()*2.0)
      action = defendGoalLine(PS->getDefendGoalLineDistance()/2.0);
    else
      action = defendGoalLine(PS->getDefendGoalLineDistance());
    ACT->putCommandInQueue(turnNeckToPoint(WM->getBallPos(), action));
    Log.logWithTime(777, "Defendline");
    //cout << "Defending goalline\n";
  }
  else
  {
    cout<<"Error in generateGoalieAction, unknown ACT_\n"<<endl;
    Log.logWithTime(777, "Error in generateGoalieAction, unknown ACT_\n");
  }

/*
  else if (mode == ACT_TELEPORT_TO_STRATEGIC_POSITION)
  {
      if (WM->isBeforeKickOff())
        formations->setFormation( FT_INITIAL );       // go to kick_off formation
      action = teleportToPos( WM->getStrategicPosition() );
      ACT->putCommandInQueue(turnNeckToPoint(WM->getBallPos(), action));
  }
  else if (mode == ACT_TURN_BODY_TO_CENTER)
  {
    ACT->putCommandInQueue(turnNeckToPoint(VecPosition(0,0), action));
    ACT->putCommandInQueue(alignNeckWithBody());
  }
*/
  adjustDashPowerForStamina(action, mode, true);
  ACT->putCommandInQueue(action);
  return true;
}



/*! This is presently based very closely on the UVA masters thesis method by
    the same name.  It presently contains a lot of bad magic numbers that should
    be calculated somehow or at least made less arbitrary.  It uses a priority
    ordering of if then else statements to determine the proper action for the
    agent to take this cycle.  Notably it doesn't plan for more than a single
    cycle, a major weakness in my view.
*/
SoccerCommand Player::determineActionWithBall()
{
  SoccerCommand action;

  /*Still have to make sure these variables are all used and reused properly*/
  VecPosition bestScoringPoint;
  VecPosition area6ThroughPassPos;
  VecPosition sideThroughPassPos;
  VecPosition outplayDestinationPos;
  VecPosition ballPos = WM->getBallPos();
  // debug
  //VecPosition target  = WM->getTargetArea();


  ObjectT     mostFreeForwardTeammate = OBJECT_ILLEGAL;
  ObjectT     teammateAtSide;
  ObjectT     mostFree456Teammate;
  ObjectT     bestTeammateInArea6;
  ObjectT     closestOpponent;
  ObjectT     agentObjectType = WM->getAgentObjectType();

  AngDeg      widestForwardDirection;
  AngDeg      widestForwardAngle;
  AngDeg      towardsGoalDirection;
  AngDeg      towardsGoalMaxAngle;
  AngDeg      towardsPenaltyAreaDirection;
  AngDeg      towardsPenaltyAreaMaxAngle;

  double      closestOpponentDistance = 0;

  closestOpponent = WM->getClosestInSetTo(OBJECT_SET_OPPONENTS, agentObjectType, &closestOpponentDistance);
  widestForwardDirection = WM->getDirectionOfWidestAngle(VecPosition(0,0), -90, 90,&widestForwardAngle, 30); //The -90 and 90 are to make it a forward angle only, the 30 is totally magical.

//if scoring probability for best scoring point > ScoringProbThr (=0.9) then
  if (WM->getBestScoringProbability(&bestScoringPoint) > PS->getShootRiskProbability())
//kick ball with maximum power to best scoring point as returned by scoring policy
  {
    cout << "Shooting!\n";
    action = kickTo(bestScoringPoint, SS->getBallSpeedMax());
  }
  else if (WM->isDeadBallUs())
  {
    mostFreeForwardTeammate = getMostFreeForwardTeammate();
    if (mostFreeForwardTeammate != OBJECT_ILLEGAL && WM->isFairlyFree(mostFreeForwardTeammate))
      action = directPass(mostFreeForwardTeammate, PASS_FAST);
    else
      action = clearBall(CLEAR_BALL_OFFENSIVE);
  }
//else if ball is located in area 1 then
  else if (WM->isBallWithinRegion(REGION_1))
//same procedure as for German Open agent; see Algorithm 9.7
// consider actions in following order (simplied conditions between brackets):
    //turn with ball towards opponents goal (if facing goal and no opponents are close)
    if (fabs(WM->getAgentGlobalBodyAngle()) <= 90 && closestOpponentDistance > PS->getOppCloseThreshold())
      action = turnWithBallTo(0,PS->getTurnWithBallAngThr(),PS->getTurnWithBallFreezeThr()); //Magic number is supposed to be towards opponents goal

    //pass ball at normal speed to most free teammate in forward direction (only if very free)
    else
    {
      mostFreeForwardTeammate = getMostFreeForwardTeammate();

      if (mostFreeForwardTeammate != OBJECT_ILLEGAL && WM->isVeryFree(mostFreeForwardTeammate))
        action = directPass(mostFreeForwardTeammate,PASS_NORMAL);

        //dribble slowly forward in direction of widest angle between opponents (if angle is wide enough)
      else if (widestForwardAngle >= 25) //Damn thats a lot of magic numbers
        action = dribble(widestForwardDirection,DRIBBLE_SLOW);

      //clear ball forward into widest angle between opponents (no condition)
      else
        action = clearBall(CLEAR_BALL_DEFENSIVE);
    }

//else if ball is located in area 2 then
  else if (WM->isBallWithinRegion(REGION_2))
  // consider actions in following order (simplied conditions between brackets):
    //turn with ball towards opponent's goal (if facing own goal and no opponents are close)
    if (fabs(WM->getAgentGlobalBodyAngle()) <= 90 && closestOpponentDistance > PS->getOppCloseThreshold())
      action = turnWithBallTo(0,PS->getTurnWithBallAngThr(),PS->getTurnWithBallFreezeThr()); //Magic number is supposed to be towards opponents goal
    else
    {
      towardsGoalDirection = WM->getDirectionOfWidestAngle(VecPosition(0,0),
      VecPosition(PITCH_LENGTH/2.0,-SS->getGoalWidth() - 5).getDirection(),
      VecPosition(PITCH_LENGTH/2.0,SS->getGoalWidth() + 5).getDirection(),
      &towardsGoalMaxAngle,
      min(WM->getAgentGlobalPosition().getDistanceTo(VecPosition(PITCH_LENGTH/2.0, 0)),30));
      //dribble fast towards opponent's goal (if no opponents are blocking the desired path)
      if (WM->getNrInSetInCone(OBJECT_SET_OPPONENTS,PS->getBlockingConeWidth(),
      ballPos, ballPos + VecPosition(PS->getFastDribbleBlockingDist(),
       towardsGoalDirection, POLAR)) <= 0)
        action = dribble(towardsGoalDirection, DRIBBLE_FAST);


      else
      {
        mostFreeForwardTeammate = getMostFreeForwardTeammate();
        //pass ball at normal speed to most free teammate in forward direction (only if very free)
        if (mostFreeForwardTeammate != OBJECT_ILLEGAL && WM->isVeryFree(mostFreeForwardTeammate))
          action = directPass(mostFreeForwardTeammate,PASS_NORMAL);

        //pass ball at high speed to most free teammate in forward direction (only if fairly free)
        else if (mostFreeForwardTeammate != OBJECT_ILLEGAL && WM->isFairlyFree(mostFreeForwardTeammate))
          action = directPass(mostFreeForwardTeammate,PASS_FAST);

        //give through pass to teammate at the side of the field (if angle between opponents is wide enough)
        else
        {
          //pick a teammate in one of the side regions, the method currently used below is crappy
          teammateAtSide = WM->getClosestInSetTo(OBJECT_SET_TEAMMATES, WM->getAgentObjectType());
          if (!WM->isObjectAlongSide(teammateAtSide))
          {
            teammateAtSide = WM->getSecondClosestInSetTo(OBJECT_SET_TEAMMATES, WM->getAgentObjectType());
            if (!WM->isObjectAlongSide(teammateAtSide))
              teammateAtSide = OBJECT_ILLEGAL;
          }

          //select the pos for the throughPass with the same y coord as the player, and an x coord 5 m short of the end of the field
          sideThroughPassPos = VecPosition(PITCH_LENGTH/2.0 - 7.5,WM->getGlobalPosition(teammateAtSide).getY()); //Oh so many magic numbers it hurts!
          //get the widest angle for a throughPass
          AngDeg widestThroughPassAngle = 0;
          SoccerCommand tempCommand = throughPass(teammateAtSide, sideThroughPassPos, &widestThroughPassAngle);
          if (teammateAtSide != OBJECT_ILLEGAL && widestThroughPassAngle > 25) //Weee! Magic numbers!
            action = tempCommand;
          //clear ball forward into widest angle between opponents (no condition)
          else
            action = clearBall(CLEAR_BALL_OFFENSIVE);
        }
      }
    }
  //else if ball is located in area 3 (i.e. 3a or 3b) then
  else if (WM->isBallWithinRegion(REGION_3a) || WM->isBallWithinRegion(REGION_3b))
  // consider actions in following order (simplied conditions between brackets):
    //turn with ball towards opponent's goal (if facing own goal and no opponents are close)
    if (fabs(WM->getAgentGlobalBodyAngle()) <= 90 && closestOpponentDistance > PS->getOppCloseThreshold())
      action = turnWithBallTo( 0, PS->getTurnWithBallAngThr(), PS->getTurnWithBallFreezeThr());

    //outplay opponent (if opponent is very close)
    else if (closestOpponentDistance <= PS->getOppVeryCloseThreshold())
      action = outplayOpponent(WM->getClosestInSetTo(OBJECT_SET_OPPONENTS, WM->getAgentObjectType()), outplayDestinationPos);

    //dribble fast towards side of opponent's penalty area (if no opponents are blocking the desired path)
    else
    {
      towardsPenaltyAreaDirection = WM->getAgentGlobalPosition().getY() <= 0 ?
         WM->getDirectionOfWidestAngle(VecPosition(0,0),VecPosition(PITCH_LENGTH/2.0,-PENALTY_AREA_WIDTH/2.0 - 5).getDirection(),VecPosition(PENALTY_X,-PENALTY_AREA_WIDTH/2.0 + 5).getDirection(),&towardsPenaltyAreaMaxAngle,min(WM->getAgentGlobalPosition().getDistanceTo(VecPosition(PITCH_LENGTH/2.0 - PENALTY_X/2.0, -PENALTY_AREA_WIDTH/2.0)),30)) :
         WM->getDirectionOfWidestAngle(VecPosition(0,0),VecPosition(PITCH_LENGTH/2.0,PENALTY_AREA_WIDTH/2.0 + 5).getDirection(),VecPosition(PENALTY_X,PENALTY_AREA_WIDTH/2.0 - 5).getDirection(),&towardsPenaltyAreaMaxAngle,min(WM->getAgentGlobalPosition().getDistanceTo(VecPosition(PITCH_LENGTH/2.0 - PENALTY_X/2.0, PENALTY_AREA_WIDTH/2.0)),30));

      if (WM->getNrInSetInCone(OBJECT_SET_OPPONENTS,PS->getBlockingConeWidth(),ballPos, ballPos + VecPosition(PS->getFastDribbleBlockingDist(), towardsPenaltyAreaDirection, POLAR)) <= 0)
        action = dribble(towardsPenaltyAreaDirection,DRIBBLE_FAST);

      //clear ball forward to my side of the field (if teammate  mostly me  can reach ball fastest)
      else if (amFastestToSideClearedBall())
        action = clearBall(CLEAR_BALL_OFFENSIVE_SIDE, (WM->getAgentGlobalPosition().getY() >= 0 ? SIDE_RIGHT : SIDE_LEFT));


      else
      {
        mostFreeForwardTeammate = getMostFreeForwardTeammate();
        //pass ball at normal speed to most free teammate in forward direction (only if very free)
        if (mostFreeForwardTeammate != OBJECT_ILLEGAL && WM->isVeryFree(mostFreeForwardTeammate))
          action = directPass(mostFreeForwardTeammate,PASS_NORMAL);

        //dribble slowly forward in direction of widest angle between opponents (if angle is wide enough)
        else if (widestForwardAngle >= 25) //So many magic numbers, so little time
          action = dribble(widestForwardDirection,DRIBBLE_SLOW);

        //clear ball forward into widest angle between opponents (no condition)
        else
          action = clearBall(CLEAR_BALL_OFFENSIVE);
      }
    }
//else if ball is located in area 4 (i.e. 4a or 4b) then
  else if (WM->isBallWithinRegion(REGION_4a) || WM->isBallWithinRegion(REGION_4b))
  {
    // consider actions in following order (simplifed conditions between brackets):
    //dribble fast towards side of opponent's penalty area (if outside penalty area and path not blocked)
    towardsPenaltyAreaDirection = WM->getAgentGlobalPosition().getY() <= 0 ?
      WM->getDirectionOfWidestAngle(VecPosition(0,0),VecPosition(PITCH_LENGTH/2.0,-PENALTY_AREA_WIDTH/2.0 - 5).getDirection(),VecPosition(PENALTY_X,-PENALTY_AREA_WIDTH/2.0 + 5).getDirection(),&towardsPenaltyAreaMaxAngle,min(WM->getAgentGlobalPosition().getDistanceTo(VecPosition(PITCH_LENGTH/2.0 - PENALTY_X/2.0, -PENALTY_AREA_WIDTH/2.0)),30)) :
      WM->getDirectionOfWidestAngle(VecPosition(0,0),VecPosition(PITCH_LENGTH/2.0,PENALTY_AREA_WIDTH/2.0 + 5).getDirection(),VecPosition(PENALTY_X,PENALTY_AREA_WIDTH/2.0 - 5).getDirection(),&towardsPenaltyAreaMaxAngle,min(WM->getAgentGlobalPosition().getDistanceTo(VecPosition(PITCH_LENGTH/2.0 - PENALTY_X/2.0, PENALTY_AREA_WIDTH/2.0)),30));
    if (!WM->isInTheirPenaltyArea(WM->getAgentGlobalPosition())  && WM->getNrInSetInCone(OBJECT_SET_OPPONENTS,PS->getBlockingConeWidth(),ballPos, ballPos + VecPosition(PS->getFastDribbleBlockingDist(), towardsPenaltyAreaDirection, POLAR)) <= 0)
      action = dribble(towardsPenaltyAreaDirection,DRIBBLE_FAST);

    //pass ball at normal speed to most free teammate in areas 4, 5 or 6 (only if very free)
    else
    {
      mostFree456Teammate = getMostFree456Teammate();
      if (mostFree456Teammate != OBJECT_ILLEGAL && WM->isVeryFree(mostFree456Teammate))
        action = directPass(mostFree456Teammate,PASS_NORMAL);

      //give through pass to teammate in area 6 (if angle between opponents is wide enough)
      else
      {
        bestTeammateInArea6 = WM->getTeammateWithinRegion(REGION_6);
        AngDeg widestThroughPassAngle = 0;
        SoccerCommand tempAction = throughPass(bestTeammateInArea6, area6ThroughPassPos, &widestThroughPassAngle);
        if (bestTeammateInArea6 != OBJECT_ILLEGAL && widestThroughPassAngle >= 25) //so many magic numbers
          action = tempAction;
        //outplay opponent (if opponent is very close)
        else if (closestOpponentDistance <= PS->getOppVeryCloseThreshold())
          action = outplayOpponent(WM->getClosestInSetTo(OBJECT_SET_OPPONENTS, WM->getAgentObjectType()), outplayDestinationPos);

        //pass ball at high speed to most free teammate in areas 4, 5 or 6 (only if fairly free)
        else if (mostFree456Teammate != OBJECT_ILLEGAL && WM->isFairlyFree(mostFree456Teammate))
          action = directPass(mostFree456Teammate,PASS_FAST);

        //clear ball towards penalty area into widest angle between opponents (no condition)
        else
          action = clearBall(CLEAR_BALL_OFFENSIVE);
      }
    }
  }
  //else if ball is located in area 5 (i.e. 5a or 5b) then
  else if (WM->isBallWithinRegion(REGION_5a) || WM->isBallWithinRegion(REGION_5b))
  {
    // consider actions in following order (simplifed conditions between brackets):
    //pass ball at normal speed to most free teammate in areas 4, 5 or 6 (only if very free)
    mostFree456Teammate = getMostFree456Teammate();
    if (mostFree456Teammate != OBJECT_ILLEGAL && WM->isVeryFree(mostFree456Teammate))
      action = directPass(mostFree456Teammate,PASS_NORMAL);

    //dribble slowly towards side of opponent's penalty area (if outside penalty area and path not blocked)
    else
    {
    towardsPenaltyAreaDirection = WM->getAgentGlobalPosition().getY() <= 0 ?
      WM->getDirectionOfWidestAngle(VecPosition(0,0),VecPosition(PITCH_LENGTH/2.0,-PENALTY_AREA_WIDTH/2.0 - 5).getDirection(),VecPosition(PENALTY_X,-PENALTY_AREA_WIDTH/2.0 + 5).getDirection(),&towardsPenaltyAreaMaxAngle,min(WM->getAgentGlobalPosition().getDistanceTo(VecPosition(PITCH_LENGTH/2.0 - PENALTY_X/2.0, -PENALTY_AREA_WIDTH/2.0)),30)) :
      WM->getDirectionOfWidestAngle(VecPosition(0,0),VecPosition(PITCH_LENGTH/2.0,PENALTY_AREA_WIDTH/2.0 + 5).getDirection(),VecPosition(PENALTY_X,PENALTY_AREA_WIDTH/2.0 - 5).getDirection(),&towardsPenaltyAreaMaxAngle,min(WM->getAgentGlobalPosition().getDistanceTo(VecPosition(PITCH_LENGTH/2.0 - PENALTY_X/2.0, PENALTY_AREA_WIDTH/2.0)),30));

      if (!WM->isInTheirPenaltyArea(WM->getAgentGlobalPosition())  && WM->getNrInSetInCone(OBJECT_SET_OPPONENTS,PS->getBlockingConeWidth(),ballPos, ballPos + VecPosition(PS->getSlowDribbleBlockingDist(), towardsPenaltyAreaDirection, POLAR)) <= 0)
        action = dribble(towardsPenaltyAreaDirection,DRIBBLE_SLOW);

      //outplay opponent (if opponent is very close)
      else if (closestOpponentDistance <= PS->getOppVeryCloseThreshold())
        action = outplayOpponent(WM->getClosestInSetTo(OBJECT_SET_OPPONENTS, WM->getAgentObjectType()), outplayDestinationPos);

      //give through pass to teammate in area 6 (if angle between opponents is wide enough)
      else
      {
        bestTeammateInArea6 = WM->getTeammateWithinRegion(REGION_6);
        AngDeg widestThroughPassAngle = 0;
        SoccerCommand tempAction = throughPass(bestTeammateInArea6, area6ThroughPassPos, &widestThroughPassAngle);
          if (bestTeammateInArea6 != OBJECT_ILLEGAL && widestThroughPassAngle >= 25) //Yay! Magic Numbers!
            action = tempAction;

        //pass ball at high speed to most free teammate in areas 4, 5 or 6 (only if fairly free)
        else if (mostFree456Teammate != OBJECT_ILLEGAL && WM->isFairlyFree(mostFree456Teammate))
          action = directPass(mostFree456Teammate,PASS_FAST);

        //clear ball towards penalty area into widest angle between opponents (no condition)
        else
          action = clearBall(CLEAR_BALL_OFFENSIVE);
      }
    }
  }
//else if ball is located in area 6 then
  else if (WM->isBallWithinRegion(REGION_6))
  {
    //// consider actions in following order (simplifed conditions between brackets):
    //dribble fast towards opponent's goal (if no opponents are blocking the desired path)
    towardsGoalDirection = WM->getDirectionOfWidestAngle(VecPosition(0,0),VecPosition(PITCH_LENGTH/2.0,-SS->getGoalWidth() - 5).getDirection(),VecPosition(PITCH_LENGTH/2.0,SS->getGoalWidth() + 5).getDirection(),&towardsGoalMaxAngle,min(WM->getAgentGlobalPosition().getDistanceTo(VecPosition(PITCH_LENGTH/2.0, 0)),30));
    if (WM->getNrInSetInCone(OBJECT_SET_OPPONENTS,PS->getBlockingConeWidth(),ballPos, ballPos + VecPosition(PS->getFastDribbleBlockingDist(), towardsGoalDirection, POLAR)) <= 0)
      action = dribble(towardsGoalDirection, DRIBBLE_FAST);

    //dribble slowly forward in direction of widest angle between opponents (if angle is wide enough)
    else if (widestForwardAngle >= 25) //Yet more magic
      action = dribble(widestForwardDirection, DRIBBLE_SLOW);

    //pass ball at normal speed to most free teammate in areas 4, 5 or 6 (only if very free)
    else
    {
      mostFree456Teammate = getMostFree456Teammate();
      if (mostFree456Teammate != OBJECT_ILLEGAL && WM->isVeryFree(mostFree456Teammate))
        action = directPass(mostFree456Teammate,PASS_NORMAL);

      //give through pass to teammate in area 6 (if angle between opponents is wide enough)
      else
      {
        bestTeammateInArea6 = WM->getTeammateWithinRegion(REGION_6);
        AngDeg widestThroughPassAngle = 0;
        SoccerCommand tempAction = throughPass(bestTeammateInArea6, area6ThroughPassPos, &widestThroughPassAngle);
        if (bestTeammateInArea6 != OBJECT_ILLEGAL && widestThroughPassAngle >= 25) //Yay magic numbers!
          action = tempAction;

        //pass ball at high speed to most free teammate in areas 4, 5 or 6 (only if fairly free)
        else if (mostFree456Teammate != OBJECT_ILLEGAL && WM->isFairlyFree(mostFree456Teammate))
          action = directPass(mostFree456Teammate,PASS_FAST);

        //kick ball with maximum power to best scoring point as returned by scoring policy (no condition)
        else
        { cout << "Secondary shooting!\n";
          action = kickTo(bestScoringPoint,SS->getBallSpeedMax());
        }
      }
    }
  }
  else
  {
    Log.logWithTime(777, "Error condition in determineActionWithBall! Unknown ACT_");
  }


//end if

  return action;
}


/*! This method is based closely on the method by the same name described in
    the UVA masters thesis. It attempts to manage stamina usage to prevent
    the agent from getting to tired and wearing itself out.

    ***possibly buggy***

    \author David Brokenshire
*/
SoccerCommand Player::adjustDashPowerForStamina(SoccerCommand action, ActionT mode, bool isGoalie)
{
  if (mode == ACT_MOVE_TO_DEAD_BALL_POSITION)
  {
    if(WM->getAgentStamina().getStamina() < SS->getStaminaMax()*3.0/4.0 )
    {
       //cout<<"ajusting dead BallPositioning"<<" stamina " <<
       //WM->getAgentStamina().getStamina()<<endl;
       action.dPower = 0;
       return action;
    }
  }


      
  if (action.commandType != CMD_DASH)
    return action; //Only dashing consumes stamina
  //Debug
  if (action.dPower > 100 || action.dPower < -100) cout << "action power messed up upon entry to adjust!\n";
  double staminaLoss = ((action.dPower > 0) ? 1 : -2) * action.dPower - SS->getStaminaIncMax(); //Ible: Should it always be SS->getStaminaIncMax()
  double diffThreshold = WM->getAgentStamina().getStamina() - SS->getRecoverDecThr() * SS->getStaminaMax();
  if (isGoalie && mode == ACT_INTERCEPT_SCORING_ATTEMPT) //if the goalie is intercepting, don't modify to save stamina
  {
    if( WM->isDeadBallUs() )
    {
       action.dPower = (action.dPower)/4.0;
    }
    return action;
  }  
  if (mode == ACT_INTERCEPT || mode == ACT_MARK)
  {
    if (diffThreshold < staminaLoss) //If this would put the agent below the threshold, change power to make the agent exactly at the threshold
    {
      action.dPower = ((action.dPower > 0) ? 1 : -0.5) * (diffThreshold + SS->getStaminaIncMax());
      if (action.dPower > 100) action.dPower = 100;
      else if (action.dPower < -100) action.dPower = -100;
    }
  }
  else if( (formations->getPlayerType() == PT_ATTACKER ||
          formations->getPlayerType() == PT_ATTACKER_WING) &&
          WM->isInDefenseArea(OBJECT_BALL)  )
  {
     action.dPower = (action.dPower)/4.0;
  }
  else if( (formations->getPlayerType() == PT_DEFENDER_SWEEPER ||
          formations->getPlayerType() == PT_DEFENDER_WING) &&
          (WM->isInOffenseArea(OBJECT_BALL) ||
          WM->getStrategicPosition().getX() +20.0 <
          WM->getBallPos().getX() ) )
  {
     action.dPower = (action.dPower)/4.0;
  }
  else
  {
    if (diffThreshold < 0.1 * SS->getStaminaMax()) //If we are within 10% of max stamina to the threshold, drop power to 0.
    {  action.dPower = 0; //save stamina
    }
    //If we are within 30 of the ball and within 25% of the threshold (1000), slow down?
    else if (diffThreshold < 0.25 * SS->getStaminaMax() && WM->getAgentGlobalPosition().getDistanceTo((WM->getBallPos())) < 30)
    {  action.dPower = (action.dPower)/4.0; //consume less stamina by dashing more slowly.
    }
    else if (diffThreshold < 0.25 * SS->getStaminaMax())
    {
       action.dPower = (action.dPower)/4.0;
    }
  }
  //Debug
  if (action.dPower > 100 || action.dPower < -100)
  {
    cout << "action power messed up upon exit from adjust!\n";
    cout << "staminaLoss = " << staminaLoss << endl;
    cout << "diffThreshold = " << diffThreshold << endl;
    cout << "Dash power = " << action.dPower << endl;
  }
  return action;
}

/*! This method allows the goalie to try to intercept the ball while staying
    on its goal defence line.  This allows it to intercept faster than it
    might otherwise.
    \author David Brokenshire
*/
SoccerCommand Player::goalieInterceptOnLine()
{
  Log.logWithTime(777, "Entered goalieInterceptOnLine()");
  //Make goalie line
  Line lineGoal = Line::makeLineFromPositionAndAngle(VecPosition( - PITCH_LENGTH/2.0 + PS->getDefendGoalLineDistance(), 0 ), 90 );
  //make ball trajectory
  Line ballTrajectory = Line::makeLineFromPositionAndAngle( WM->getBallPos(), WM->getBallDirection());
  //find intersection (get z)
	// ~z = desired guard point, t + t = agent's global body angle in cycle t
	//~z = intersection(ball trajectory, goalie line)
  VecPosition z = lineGoal.getIntersection(ballTrajectory);

  //Find the correct angle to make goalie look in direction of ball when standing on interception point.
  //AngDeg desiredGlobalBodyDir = (WM->getBallPos() - z).getDirection(); //I believe this is the correct global body angle
  //return moveToPosAlongLine(~z, t + t, GoalieDevDist, 1, GoalieTurnThr, GoalieCorrAng)
  int         iSol;
  VecPosition posBall = getInterceptionPointBall( &iSol, true, INTERCEPTION_AS_SECOND   );

  if (posBall.getDistanceTo(WM->getPosOwnGoal()) > z.getDistanceTo(WM->getPosOwnGoal()) &&
      posBall.getX() > -PITCH_LENGTH/2.0 )
    z=posBall;

  if( WM->getAgentGlobalPosition().getDistanceTo(z) > SS->getCatchableAreaW() - 0.1 )
    return moveToPos(z,20);
  else
    return turnBodyToObject(OBJECT_BALL);
  //return moveToPosAlongLine(z, desiredGlobalBodyDir, 3.0, 1, 7, 12); //These magic numbers should be put into player settings
}


/*! This method allows the goalie to try to intercept a cross pass.
    it is used with WM->isCrossPass() only.

    \author Serguei Razykov
*/
SoccerCommand Player::goalieInterceptCrossPass()
{
  Log.logWithTime(777, "Entered goalieInterceptCrossPass()");
  //Make defend line
  Line lineGoal = Line::makeLineFromPositionAndAngle(WM->getAgentGlobalPosition(), 0.0 );
  //make ball trajectory
  Line ballTrajectory = Line::makeLineFromPositionAndAngle( WM->getBallPos(), WM->getBallDirection());
  //find intersection (get z)
	// ~z = desired guard point, t + t = agent's global body angle in cycle t
	//~z = intersection(ball trajectory, goalie line)
  VecPosition z = lineGoal.getIntersection(ballTrajectory);

  //Find the correct angle to make goalie look in direction of ball when standing on interception point.
  //AngDeg desiredGlobalBodyDir = (WM->getBallPos() - z).getDirection(); //I believe this is the correct global body angle
  //return moveToPosAlongLine(~z, t + t, GoalieDevDist, 1, GoalieTurnThr, GoalieCorrAng)
  /*int         iSol;
  VecPosition posBall = getInterceptionPointBall( &iSol, true  );

  if (posBall.getDistanceTo(WM->getAgentGlobalPosition()) > z.getDistanceTo(WM->getAgentGlobalPosition()) )
    z=posBall; */

  if( WM->getAgentGlobalPosition().getDistanceTo(z) > SS->getCatchableAreaW() - 0.1 )
    return moveToPos(z,20);
  else
    return turnBodyToObject(OBJECT_BALL);
  //return moveToPosAlongLine(z, desiredGlobalBodyDir, 3.0, 1, 7, 12); //These magic numbers should be put into player settings
}


/*! This method gets a vector of teammates ordered by their freedom score and
    then finds the one with a x coord greater than that of the agent and the
    highest freedom score
    \return ObjectT of the forward teammate with the highest freedom score
    \author David Brokenshire
*/
ObjectT Player::getMostFreeForwardTeammate()
{
  VecPosition posAgent = WM->getAgentGlobalPosition();
  vector<ObjectT> freeTeammates = WM->getFreeTeammateVector();      //freedom list is inorder
  //find the first agent in the forward direction (x coord > our agent's x-coord)
  for (int i = 0; i < (int)freeTeammates.size(); ++i)
  {
    if (WM->getGlobalPosition(freeTeammates[i]).getX() > posAgent.getX())
    {
      return freeTeammates[i];//We want the most free, which is the closest to the front of the vector, so get us out
    }
  }
  return OBJECT_ILLEGAL;
}

/*This method maps free teammates by directions
*/

void Player::mapFreeTeammates(ObjectT *leftBack,ObjectT *left,
                              ObjectT *leftForward, ObjectT *forward,
                              ObjectT *rightForward,
                              ObjectT *right, ObjectT *rightBack,
                              ObjectT *back, ObjectT *reg6, ObjectT *reg6a,
                              VecPosition pos)
{
  *leftBack = OBJECT_ILLEGAL;
  *left = OBJECT_ILLEGAL;
  *leftForward = OBJECT_ILLEGAL;
  *forward = OBJECT_ILLEGAL;
  *rightForward = OBJECT_ILLEGAL;
  *right = OBJECT_ILLEGAL;
  *rightBack = OBJECT_ILLEGAL;
  *back = OBJECT_ILLEGAL;
  *reg6 = OBJECT_ILLEGAL;
  *reg6a = OBJECT_ILLEGAL;
  vector<ObjectT> freeTeammates = WM->getFreeTeammateVector();      //freedom list is in order
  AngDeg direction;
//  VecPosition pen = VecPosition(PITCH_LENGTH/2.0+11.0, 0.0);

  //find the first agent in the each direction ( in definite angle)
  for (int i = 0; i < (int)freeTeammates.size(); ++i)
  {
    // debug
    //if (WM->getBallPos().getX() > 30)
      //cout<<"# "<<i<<" pos "<<WM->getGlobalPosition(freeTeammates[i])<<endl;
    // we don't want to pass in offside
    if (WM->getGlobalPosition(freeTeammates[i]).getX() > WM->getOffsideX())
      continue;

    direction = (WM->getGlobalPosition(freeTeammates[i])- pos).getDirection();
    direction = VecPosition::normalizeAngle( direction);
    // most left
    if (direction < -150.0)
    {
      if (*back == OBJECT_ILLEGAL)
       *back = freeTeammates[i];//We want the most free
    }
    else if (direction < -95.0)
    {
      if (*leftBack == OBJECT_ILLEGAL)
       *leftBack = freeTeammates[i];//We want the most free
    }
    else if ( direction < - 75.0)
    {
      if (*left == OBJECT_ILLEGAL)
       *left = freeTeammates[i];//We want the most free
    }
    // left
    else if (direction < - 30.0)
    {
      if (*leftForward == OBJECT_ILLEGAL)
       *leftForward = freeTeammates[i];//We want the most free
    }
    // forward
    else if (direction <  30.0)
    {
      if (*forward == OBJECT_ILLEGAL)
       *forward = freeTeammates[i];//We want the most free
    }
    // right fotward
    else if (direction <  75.0)
    {
      if (*rightForward == OBJECT_ILLEGAL)
       *rightForward = freeTeammates[i];//We want the most free
    }
    // right
    else if (direction < 95.0)
    {
      if (*right == OBJECT_ILLEGAL)
       *right = freeTeammates[i];//We want the most free
    }
    else if (direction < 150.0)
    {
      if (*rightBack == OBJECT_ILLEGAL)
       *rightBack = freeTeammates[i];//We want the most free
    }
    else
    {
      if (*back == OBJECT_ILLEGAL)
       *back = freeTeammates[i];//We want the most free
    }
    // in reg 6a
    if (WM->isObjectWithinRegion(freeTeammates[i], REGION_6)  &&
        (*reg6 == OBJECT_ILLEGAL ))
    {
     // cout<<" in 6 - #"<<i<<endl;
      *reg6 = freeTeammates[i];
    }

    // in reg 6a
    if ( WM->isObjectWithinRegion(freeTeammates[i], REGION_6a) &&
        (*reg6a == OBJECT_ILLEGAL))
    {
     // cout<<" in 6 - #"<<i<<endl;
      *reg6a = freeTeammates[i];
    }
  }
  return ;
}

/*! This method is undocumented...
*/
ObjectT Player::getMostFreeForwardTeammateP(VecPosition point)
{
  VecPosition posAgent = WM->getAgentGlobalPosition();
  vector<ObjectT> freeTeammates = WM->getFreeTeammateVectorP(point);      //freedom list is inorder
  //find the first agent in the forward direction (x coord > our agent's x-coord)
  for (int i = 0; i < (int)freeTeammates.size(); ++i)
  {
    if (WM->getGlobalPosition(freeTeammates[i]).getX() > posAgent.getX())
    {
      return freeTeammates[i];//We want the most free, which is the closest to the front of the vector, so get us out
    }
  }
  return OBJECT_ILLEGAL;
}
/*! This method returns the ObjectT of the most free teammate who is in region
    4a,4b,5a,5b, or 6.

    ***incomplete***

    \return ObjectT of the most free player in regions 4,5, and 6
    \author David Brokenshire
*/
ObjectT Player::getMostFree456Teammate()
{
  vector<ObjectT> freeTeammates = WM->getFreeTeammateVector();      //freedom list is inorder
  //find the first agent in the forward direction (x coord > our agent's x-coord)
  for (int i = 0; i < (int)freeTeammates.size(); ++i)
  {
    if (
        /*WM->isObjectWithinRegion(freeTeammates[i],REGION_4a) ||
        WM->isObjectWithinRegion(freeTeammates[i],REGION_4b) ||
        WM->isObjectWithinRegion(freeTeammates[i],REGION_5a) ||
        WM->isObjectWithinRegion(freeTeammates[i],REGION_5b) ||*/
        WM->isObjectWithinRegion(freeTeammates[i],REGION_6))
    {
      return freeTeammates[i];//We want the most free, which is the closest to the front of the vector, so get us out
    }
  }
  return OBJECT_ILLEGAL; //If we don't have any free teammates in those regions return OBJECT_ILLEGAL
}

/*! This method returns the ObjectT of the most free teammate who is in circle around
    the agent with radius "radius"

    \return ObjectT of the most free player in circle
    \author David Brokenshire
*/
ObjectT Player::getMostFreeAroundTeammate(double radius)
{
  vector<ObjectT> freeTeammates = WM->getFreeTeammateVector();      //freedom list is inorder
  //find the first agent in the forward direction (x coord > our agent's x-coord)
  Circle around = Circle(WM->getAgentGlobalPosition(), radius);
  for (int i = 0; i < (int)freeTeammates.size(); ++i)
  {
    if ( around.isInside(WM->getGlobalPosition(freeTeammates[i])) )
    {
      if (freeTeammates[i]==OBJECT_TEAMMATE_1 )
        continue;
      return freeTeammates[i];//We want the most free, which is inside the circle
    }
  }
  return OBJECT_ILLEGAL; //If we don't have any free teammates in those regions return OBJECT_ILLEGAL
}


/*! This method returns the ObjectT of the most free teammate who is in target area

*/
ObjectT Player::getMostFreeTargetTeammate()
{
  vector<ObjectT> freeTeammates = WM->getFreeTeammateVector();      //freedom list is inorder
  //find the first agent in the forward direction (x coord > our agent's x-coord)
  for (int i = 0; i < (int)freeTeammates.size(); ++i)
  {
    if (WM->isObjectWithinRegion(freeTeammates[i],REGION_T))

    {
      return freeTeammates[i];//We want the most free, which is in the target area
    }
  }
  return OBJECT_ILLEGAL; //If we don't have any free teammates in those regions return OBJECT_ILLEGAL
}

/*! This method returns the ObjectT of the most free teammate inside the rectangle rect

*/
ObjectT Player::getMostFreeRectTeammate(Rectangle rect)
{
  vector<ObjectT> freeTeammates = WM->getFreeTeammateVector();      //freedom list is inorder
  //find the first agent in the forward direction (x coord > our agent's x-coord)
  for (int i = 0; i < (int)freeTeammates.size(); ++i)
  {
    if (rect.isInside(WM->getGlobalPosition(freeTeammates[i])))

    {
      return freeTeammates[i];//We want the most free, which is in the rectangle
    }
  }
  return OBJECT_ILLEGAL; //If we don't have any free teammates in those regions return OBJECT_ILLEGAL
}

/* This method defines if the agent should intercept. Created by Serguei.

*/

bool Player::shouldIIntercept( InterceptionT *iType )
{
  bool should = false;
  
  TiredNessT tired = WM->getAgentTiredNess();
  VecPosition posBall = WM->getBallPos();
  // fastest player and number of cycles
  int iFastest;
  ObjectT fastestPlayer;

  int iOpponent =100;

  ObjectT fastestOpponent = WM->getFastestInSetTo(OBJECT_SET_OPPONENTS, OBJECT_BALL, &iOpponent);
  // fastest teammate and number of cycles
  int iTeammate = 100;
  ObjectT fastestTeammate = WM->getFastestInSetTo(OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTeammate);
  // second fastest teammate and number of cycles
  int iTeammate1 = 100;
  ObjectT secondFastestTeammate = WM->getSecondFastestInSetTo(OBJECT_SET_TEAMMATES, OBJECT_BALL, &iTeammate1);

  PlayerT fastestTeammateType = formations->getPlayerType(SoccerTypes::getIndex(fastestTeammate));
  PlayerT secondFastestTeammateType = formations->getPlayerType(SoccerTypes::getIndex(secondFastestTeammate));


  if(iOpponent < iTeammate)
  {
     fastestPlayer = fastestOpponent;
     iFastest = iOpponent;
  }
  else
  {
     fastestPlayer = fastestTeammate;
     iFastest = iTeammate;
  }
  // If I am the fastest player I must intercept except it is goalKickUs
  if(fastestPlayer == WM->getAgentObjectType() && !WM->isGoalKickUs())
  {
    should = true;
  }
  else if(fastestTeammate == WM->getAgentObjectType()) // I am the fastest teammate
  {
     //I must untercept except...
     should = true;

     // goal Kick Us
     if (WM->isGoalKickUs())
     {
        should = false;
     }
     else if (WM->getTimeSinceLastCatch() < 20 )
     {
        should = false;
     }
      // an opponent much closer to the ball
     else if( iTeammate > iFastest + 35 && !WM->isDeadBallUs())  // extermely magic number. Must be derived from an experiment
                                     // and placed to PlayerSettings
     {
      // cout<<"an opponent too far"<<endl;
       should = false;
     }

      //I am defender and the ball is on the opponent's side or...
     else if( ((WM->getPlayerNumber() < 6 && posBall.getX() > 0 ) ||
              fabs(WM->getStrategicPosition().getY() - posBall.getY() ) > 12.0 &&
              !WM->isInDefenseArea(OBJECT_BALL)) && !WM->isDeadBallUs())
     {
       should = false;
     }


     /* I am defender and the ball is in the corner and an opponent is closer to the ball
     else if(WM->getPlayerNumber() < 6 && posBall.getX() < -PITCH_LENGTH/2.0 + 20.0 &&
              fabs(posBall.getY()) > 22.0 && !WM->isDeadBallUs() && iTeammate > iFastest + 5)
     {
       should = false;
     }*/

     // set a style of intercepting
     // if I am a midfield and a second fastest is a defender and close enough
     if( (fastestTeammateType == PT_MIDFIELDER_CENTER ||
       fastestTeammateType == PT_MIDFIELDER_WING) &&
       (secondFastestTeammateType == PT_DEFENDER_SWEEPER ||
        secondFastestTeammateType == PT_DEFENDER_WING) )
     {
       *iType = INTERCEPTION_AS_SECOND;
     }
     // I am attacker, an attacker always intersepts as the forward
     else if( fastestTeammateType == PT_ATTACKER ||
       fastestTeammateType == PT_ATTACKER_WING )
     {
       *iType = INTERCEPTION_AS_FORWARD;
     }
     else if( fastestTeammateType == PT_DEFENDER_SWEEPER ||
       fastestTeammateType == PT_DEFENDER_WING )
     {
       *iType = INTERCEPTION_AS_DEFENDER;
     }
     else if( fastestTeammateType == PT_MIDFIELDER_CENTER ||
       fastestTeammateType == PT_MIDFIELDER_WING )
     {
       *iType = INTERCEPTION_AS_FORWARD;
     }
  }
  // I am the second fastest and the fastest teammate is not the fastest player
  else if (secondFastestTeammate == WM->getAgentObjectType() && fastestTeammate != fastestPlayer)
  {
    should = false;
    if(SoccerTypes::isGoalie(fastestTeammate)    )
    {
      should = true;
      *iType = INTERCEPTION_AS_DEFENDER;
    }
    // I am too tired
    else if (tired == TIREDNESS_BAD )
    {
      should = false;
    }
    // let the fastest to intercept
    else if (WM->isDeadBallUs() )
    {
      should = false;
    }
    // I am a midfielder and the fastest is not midfielder
    else if( (secondFastestTeammateType == PT_MIDFIELDER_CENTER ||
      secondFastestTeammateType == PT_MIDFIELDER_WING) &&
       (fastestTeammateType != PT_MIDFIELDER_CENTER &&
      fastestTeammateType != PT_MIDFIELDER_WING))
    {
       should = true;
       *iType = INTERCEPTION_AS_SECOND;
    }
    // if the fastest teammate is behind the opponent and I am not
   // else if( fabs( (posBall - WM->getGlobalPosition(fastestOpponent)).getDirection() -
   //                (WM->getGlobalPosition(fastestTeammate) -
   //                WM->getGlobalPosition(fastestOpponent)).getDirection() ) > 90.0 &&
   //                fabs((posBall - WM->getGlobalPosition(fastestOpponent)).getDirection() -
   //                (WM->getPosOwnGoal() -
   //                WM->getGlobalPosition(fastestOpponent)).getDirection() ) < 90.0) // magic number

    else if( fabs( (WM->getPosOwnGoal() - WM->getGlobalPosition(fastestOpponent)).getDirection() -
                   (WM->getGlobalPosition(fastestTeammate) -
                   WM->getGlobalPosition(fastestOpponent)).getDirection() ) > 105.0 &&
                   fabs((posBall - WM->getGlobalPosition(fastestOpponent)).getDirection() -
                   (WM->getPosOwnGoal() -
                   WM->getGlobalPosition(fastestOpponent)).getDirection() ) < 75.0) // magic number
    {
       should = true;
       *iType = INTERCEPTION_AS_DEFENDER;
       if (fabs(WM->getStrategicPosition().getY() - posBall.getY() ) > 12.0 ||
                secondFastestTeammateType == PT_DEFENDER_SWEEPER ||
               secondFastestTeammateType == PT_DEFENDER_WING)
       {
         should = false;
       }
    }

   // else if(WM->getGlobalPosition(fastestOpponent).getX() <  WM->getGlobalPosition(fastestTeammate).getX())
    /* if the fastest teammate is behind the opponent and I am not
    else if( !WM->isInDefenseArea(OBJECT_BALL) &&
              WM->getGlobalPosition(fastestPlayer).getX() <
              WM->getGlobalPosition(fastestTeammate).getX() + 0.5 &&
              (fastestTeammateType == PT_DEFENDER_SWEEPER ||
               fastestTeammateType == PT_DEFENDER_WING))

    {
       should = true;
       *iType = INTERCEPTION_AS_SECOND;
    } */
    /*if the fastest teammate is behind the opponent and I am not
    else if( (secondFastestTeammateType == PT_DEFENDER_SWEEPER ||
      secondFastestTeammateType == PT_DEFENDER_WING) &&
      (fastestTeammateType == PT_DEFENDER_SWEEPER ||
      fastestTeammateType == PT_DEFENDER_WING)
      && WM->isInDefenseArea(OBJECT_BALL) )//&&
          //   WM->getGlobalPosition(fastestOpponent).getDistanceTo(WM->getPosOwnGoal()) < 20.0 )
     //              WM->getGlobalPosition(fastestTeammate).getDistanceTo(WM->getPosOwnGoal())+1.5 )   // magic number
    {

       should = true;
       *iType = INTERCEPTION_AS_DEFENDER;
    }*/

    // I am a defender, close enough and the fastest teammate is a midfield
    else if( (secondFastestTeammateType == PT_DEFENDER_SWEEPER ||
      secondFastestTeammateType == PT_DEFENDER_WING) &&
       (fastestTeammateType == PT_MIDFIELDER_CENTER ||
      fastestTeammateType == PT_MIDFIELDER_WING)  )
    {
       should = true;
       // a defender always intercepts as a defender
       *iType = INTERCEPTION_AS_DEFENDER;
    }
    // I am a attacker, close enough and the fastest teammate is a midfield
    else if( (secondFastestTeammateType == PT_ATTACKER ||
      secondFastestTeammateType == PT_ATTACKER_WING) &&
       (fastestTeammateType == PT_MIDFIELDER_CENTER ||
      fastestTeammateType == PT_MIDFIELDER_WING) )
    {
       should = true;
       // an attacker always intercepts as a second
       *iType = INTERCEPTION_AS_SECOND;
    }

  }

  return should;

}
/* This method works with hideBallMultyCycle.
The agent must hide the ball if the opponent "o" is close enough and the angle
between it and the ball is less than 90 degrees.
/author S.Razykov
*/
bool Player::shouldIHideBall(ObjectT o)
{
  bool hide = false;
  AngDeg angOpp  = WM->getRelativeAngle(o, true);
  AngDeg angBall = WM->getRelativeAngle(OBJECT_BALL, true);
  AngDeg angDiff = fabs( VecPosition::normalizeAngle( angOpp - angBall ) );
  //double dDist = WM->getRelativeDistance(o);
  double dDist = WM->getGlobalPosition(o).getDistanceTo(WM->getBallPos());

  if (angDiff < 90.0 && dDist < PS->getOppVeryCloseThreshold() )
  {
    hide = true;
  }

  return hide;    
  
}

bool Player::shouldIScore( VecPosition *pos)
{
  bool shoot = false;
  VecPosition posBall = WM->getBallPos();
  VecPosition posGoalie = WM->getOpponentGoaliePos();
  VecPosition leftStick = VecPosition(PITCH_LENGTH/2.0, -6.5);
  VecPosition rightStick = VecPosition(PITCH_LENGTH/2.0, 6.5);
  //VecPosition bestScoringPoint;

//  AngDeg      scoringDirection;
  AngDeg      scoringMaxAngle;
  AngDeg      goalieRelDir = (posGoalie-posBall).getDirection();
  AngDeg      leftStickRelDir = (leftStick - posBall).getDirection();
  AngDeg      rightStickRelDir = (rightStick - posBall).getDirection();

  

  if ( goalieRelDir - leftStickRelDir > rightStickRelDir - goalieRelDir )
  {
    scoringMaxAngle = goalieRelDir - leftStickRelDir;
    *pos = leftStick;
    //scoringDirection = leftStickRelDir+1.0;

  }
  else
  {
    scoringMaxAngle = rightStickRelDir - goalieRelDir;
    //scoringDirection = rightStickRelDir-1.0;
    *pos = rightStick;
  }

  
  if (WM->getOppGoalieType() != OBJECT_ILLEGAL &&
      ( goalieRelDir < leftStickRelDir && goalieRelDir < rightStickRelDir ||
        goalieRelDir > leftStickRelDir && goalieRelDir > rightStickRelDir ) &&
         scoringMaxAngle > 25.0 )
  {
     cout<<"shoot to free goal"<<endl;
     shoot = true;
     *pos = WM->getPointBetweenTwoClosestOpponentsInCone(posBall,leftStick,rightStick, 20.0);
  }
  else if (WM->getOppGoalieType() != OBJECT_ILLEGAL && scoringMaxAngle > 60.0   )
  {
     cout<<"shoot for wide angle "<<endl;
     shoot = true;
     *pos = WM->getPointBetweenTwoClosestOpponentsInCone(posBall,leftStick,rightStick, 20.0);
  }
  else if (posBall.getDistanceTo(WM->getPosOpponentGoal()) > 24.0 )
  {
    shoot = false;
  }
  else if ( scoringMaxAngle > 35.0 )
  {
     cout<<"shoot by angle"<<endl;
     shoot = true;
     //action = ACT_KICK_TO;
     //pointToPass = posBall + VecPosition(25.0, scoringDirection, POLAR);
  }
  else if ( posBall.getDistanceTo(posGoalie) < 6.0 )
  {
     cout<<"shoot by distance to goalie"<<endl;
    //action = ACT_KICK_TO;
    if (WM->getAgentGlobalPosition().getY() < 0.0)
       *pos = leftStick;
    else
       *pos = rightStick;

    shoot = true;
  }
  else if( posBall.getDistanceTo(WM->getPosOpponentGoal() ) < 11.0 )
           //&& posBall.getX() < PITCH_LENGTH/2.0 -4.0 )
  {
    cout<<"shoot by distance"<<endl;
    if (WM->getAgentGlobalPosition().getY() < 0.0)
       *pos = rightStick;
    else
       *pos = leftStick;

    shoot = true;
  }

  return shoot;
}

/*! This method determines whether or not the agent should mark a player. This
    method creates a list containing all the opponents that are located in the
    agent's defensive quarter of the field (i.e. the x-coord of their position
    is < -pitch_length/4) and assigns a priority factor to each of these
    opponents which represents the importance of marking it. The priority factor
    depends on the distance from the opponent to the penalty spot and on whether
    the opponent is currently located on the same side of the field as the ball.
    The priority factor fi which is assigned to an opponent i is calculated
    according to the following formula:

    fi = (1 + 2*theta(sign(pty)=sign(qty)))/magnitude((PENALTY_SPOT_X, 0)- vec(Pt))

    theta returns 1 if a boolean is true, zero otherwise.
    pty is a vector representing the current position of the opponent
    qty is a vector representing the current position of the ball

    The list of opponents is then sorted based on this factor and for each opponent
    on the list it is determined which teammate should be his marker. The highest
    priority opponent is assigned to the teammate who is closest to him, the second
    highest to the closest remaining teammate and so on.  This mapping can be used
    by the agent to determine if he should mark an opponent.  If we do determine
    that an opponent should be marked, we are also responsible for recording which
    opponent in this method.

    Sets the internal player variable for which player should be marked to the appropriate
    ObjectT. This seems kind of side-effecty and not that good

    \author David Brokenshire

    ***possibly buggy***
*/
bool Player::shouldIMark()
{
//  int Cycle = WM->getCurrentCycle();
//  bool should;
  vector<ObjectT> opponentsInDefensiveQuarter;
  Circle defenseRound = Circle(VecPosition(-PITCH_LENGTH/2, 0.0), 32);
  int i = 0;
  VecPosition opponentPos;
  double previousPriority = -1000;
  double currentPriority = -1000;
  vector<ObjectT>::iterator lastAddition;
  //Create list of opponents in our defensive quarter
  //Calculate and assign priority factors to each opponent
  //sort list of opponents based on priority.
  for (ObjectT object = WM->iterateObjectStart(i, OBJECT_SET_OPPONENTS); object != OBJECT_ILLEGAL; object = WM->iterateObjectNext(i, OBJECT_SET_OPPONENTS))
  {
    opponentPos =WM->getGlobalPosition(object);
    //if the opponent is in the defensive quarter
    //if (opponentPos.getX() < -PITCH_LENGTH/4.0)
    if (defenseRound.isInside(opponentPos) )
    {
      previousPriority = currentPriority;
      //calculate priority factors
      currentPriority = calculatePriority(opponentPos);
      //insert into list in priority order
      if (opponentsInDefensiveQuarter.size() == 0)
      {
        opponentsInDefensiveQuarter.push_back(object);
        lastAddition = opponentsInDefensiveQuarter.begin();
      }
      else if (currentPriority >= previousPriority)
        lastAddition = opponentsInDefensiveQuarter.insert(lastAddition, object);
      else
        lastAddition = opponentsInDefensiveQuarter.insert(lastAddition+1, object);
    }
  }
  // if no opponents in area - return false
  if ( opponentsInDefensiveQuarter.size()<1 )
  {
    opponentToMark = OBJECT_ILLEGAL;
   //cout<<"no opponents to mark"<<endl;

    return false;
  }

  //Make a list of available teammates
  i = 0;
  vector<ObjectT> teammates;
  for (ObjectT object = WM->iterateObjectStart(i, OBJECT_SET_TEAMMATES); object != OBJECT_ILLEGAL; object = WM->iterateObjectNext(i, OBJECT_SET_TEAMMATES))
  {
    //if (object == OBJECT_TEAMMATE_1||object == OBJECT_TEAMMATE_GOALIE)
      //continue;
    teammates.push_back(object);
  }

  //for each opponent in the list
  int numberOfOpp = opponentsInDefensiveQuarter.size();
  double minimumDistance = 1000;
  int minimumIndex = -1000;
  double newDistance = 1000;
  vector<ObjectT>::iterator minimumIterator;
  ObjectT selectedTeammate = OBJECT_ILLEGAL;
  for (int j = 0; j < numberOfOpp; ++j)
  {
    minimumDistance = 1000;
    //select the closest available teammate as his marker
    minimumIterator = teammates.begin();
    for (int i = 0; i < (int)teammates.size(); ++i)
    {
      //get the distance from the ith teammate to the jth opponent
      newDistance = WM->getGlobalPosition(teammates[i]).getDistanceTo(WM->getGlobalPosition(opponentsInDefensiveQuarter[j]));
      //if this distance is the shortest to the jth opponent for any teammate so far record it
      if (newDistance < minimumDistance)
      {
        minimumDistance = newDistance;
        minimumIndex = i;
        minimumIterator++;
      }
    }
    //erase teammate from list of available
    selectedTeammate = teammates[minimumIndex];
    teammates.erase(minimumIterator);
    //if the agent has been selected to mark someone --Note that this assumes that if there is an opponent for us to mark, we should always mark him.
    if (WM->getAgentObjectType() == selectedTeammate)
    {
      //record which opponent in a player variable
      opponentToMark = opponentsInDefensiveQuarter[j];
      //return true

      return true;
    }
  }
  opponentToMark = OBJECT_ILLEGAL;
  //cout<<"no opponents to mark"<<endl;

  return false;
}

/*! This method uses the formula described in the UVA masters thesis to
    determine a priority rating for opponents to be marked.

    \param opponentPos The position of an opponent player whose priority we want
    \return double The priority assigned to the opponent
    \author David Brokenshire
    \author S.Razykov
*/
double Player::calculatePriority(const VecPosition& opponentPos)
{
  return (1 + 2*(sign(opponentPos.getY()) == sign(WM->getBallPos().getY()) ? 1 : 0))/(VecPosition(PENALTY_X, 0) - opponentPos).getMagnitude();
  // let's try to mark those who is closest to our goal S.Razykov
  //return 1/(VecPosition(PENALTY_X, 0) - opponentPos).getMagnitude();
}

bool Player::shouldIMark1(VecPosition pos)
{
  //
  bool mark = false;
  double dDist;
  double dDist1;
  
  
  ObjectT opponent =  WM->getClosestRelativeInSet(OBJECT_SET_OPPONENTS, &dDist);
  ObjectT teammate =  WM->getClosestInSetTo(OBJECT_SET_TEAMMATES, opponent, &dDist1);

  
  // if the opponent is in penalty area
  if( WM->getGlobalPosition(opponent).getX() < -PITCH_LENGTH/2.0 + 18.0 &&
      fabs( WM->getGlobalPosition(opponent).getY() ) < 22.0 )// ||
      //WM->isInDefenseArea(opponent) )
      //&&WM->getBallPossesion() != POSSESION_US)
  //if( WM->isInOwnPenaltyArea(opponent) )
  {
    //cout<<"I am cheking for mark"<<endl;
    //if( WM->getBallPossesion() != POSSESION_US )
    //{
      // cout<<"I am cheking for mark "<<WM->getPlayerNumber()<<endl; 
      // it is close enough to my strategy position
      if (WM->getStrategicPosition().getDistanceTo(WM->getGlobalPosition(opponent)) < 12.0)
      {
      // I am closest to the opponent
        if(teammate == WM->getAgentObjectType() )
        {
          opponentToMark = opponent;
          mark = true;
        }
        // I am the second closest and the first must intercept or is the goalie
        else if( (teammate == WM->getFastestInSetTo(OBJECT_SET_TEAMMATES, OBJECT_BALL)||
                teammate == WM->getOwnGoalieType()) &&
                WM->getSecondClosestInSetTo(OBJECT_SET_TEAMMATES, opponent) ==
                WM->getAgentObjectType()  )
        {
           opponentToMark = opponent;
           mark = true;
        }
      }
    //}
  }

  return mark;
 
}

/*! This method checks if a teammate is the fastest player to
    a ball that is cleared along the side of the field.  It
    allows the player to clear the ball up field and then chase
    the ball down before any other player can get to it.

    \return bool true if this player is fastest to ball.
    \author David Brokenshire
    ***incomplete***
*/
bool Player::amFastestToSideClearedBall()
{
  //Stub
  cout << "called stub for amFastestToSideClearedBall\n";
  return false;

  //figure out where the ball will be at the end of a side clear
  //if the fastest player to that spot in OBJECT_SET_PLAYERS is a teammate
  //we are the fastest to the side cleared ball

  //Do we have a method that figures out ball end speed?
}


/*! This method deals with setting up the player in the correct
    location and positioning for the kickoff.
*/
void Player::setUpForKickOff()
{
  if( formations->getFormation() != FT_INITIAL )
    formations->setFormation( FT_INITIAL );       // go to kick_off formation
  VecPosition posStrat = WM->getStrategicPosition();
  if( WM->isKickOffThem( ) && posStrat.getDistanceTo( VecPosition(0,0) ) < 9.0 )
    posStrat.setX( -10.0 );

  if( WM->getAgentGlobalPosition().getDistanceTo( posStrat ) > 2.0  )
    ACT->putCommandInQueue( teleportToPos( posStrat ) );
  else                                            // else turn to center
  {
    ACT->putCommandInQueue( turnBodyToPoint( VecPosition( 0, 0 ), 0 ) );
    ACT->putCommandInQueue( alignNeckWithBody( ) );
  }
}

// an implementation of the first multicycle skill. It allow the player to dribble the ball
// for a some time but only one intercept after kick is perfomed.
// still requires some extra work
LevelStateT Player::dribbleSimpleMultiCycle( AngDeg ang, DribbleT d, VecPosition p)
{
  SoccerCommand soc = SoccerCommand (CMD_ILLEGAL);
  LevelStateT   lst;
  double d1;
  VecPosition pos;
  InterceptionT iType;

  if (d == DRIBBLE_SLOW)
  {
    d1 = PS->getSlowDribbleBlockingDist();
  }
  else if(d == DRIBBLE_FAST)
  {
    d1 = PS->getFastDribbleBlockingDist();
  }
  else
  {
    d1 = 2.0;
  }
  //cout<<"dribble?\n";
  if (!WM->isConfidenceGood(OBJECT_BALL))
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
    return lst;
  }

  if ( WM->isBallKickable() )
  {
    //stop dribble if can score
    if (shouldIScore(&pos))
    {
      lst =  ACTIONS_UNSELECTED_FINISHED;
      return lst;
    }
    // stop dribble if distance covered
    if (WM->getBallPos().getDistanceTo(p) > d1)
    {
      lst =  ACTIONS_UNSELECTED_FINISHED;
      return lst;
    }
    
    
    // stop dribble if the situation was changed
    if ( !WM->canItoDribble(ang,  d)  )
    {
      lst =  ACTIONS_UNSELECTED_FINISHED;
      return lst;
    } 
    /*
    if (WM->getNrInSetInCone(OBJECT_SET_OPPONENTS, 0.3, WM->getBallPos(),
      WM->getBallPos() + VecPosition(d1 , ang, POLAR)) > 0 )
    {
      lst =  ACTIONS_UNSELECTED_FINISHED;
      return lst;
    } */
    
    if ( multiCycleDribbleStep == 0 ) // first stage of the dribble
    {
      soc = dribble( ang, d);
      if (soc.commandType == CMD_ILLEGAL)        // action failed
      {
        soc = searchBall();
        lst = ACTIONS_SELECTED_FINISHED;
        multiCycleDribbleStep = 0;  // recover initial value
         }
      else
      {
        lst = ACTIONS_SELECTED;
      }
    }
    else // the ball is kickable after an interception, the dribble is finished succesfully
    {
      if (soc.commandType != CMD_ILLEGAL)
      {
          ACT->putCommandInQueue( soc );
          ACT->putCommandInQueue(turnNeckDiffSide( soc));
      }
      multiCycleDribbleStep = 0;    // recover initial value
      lst = ACTIONS_UNSELECTED_FINISHED;
    }
  }
  else  // the ball is not kickable
  {
    // if the agent is still fastest to the ball
    if (shouldIIntercept( &iType))
    {
      //when dribble always intersept as the second
      soc = intercept(false, INTERCEPTION_AS_SECOND);
      if (soc.commandType == CMD_ILLEGAL)        // action failed
      {
        lst = ACTIONS_UNSELECTED_FAILED;
        multiCycleDribbleStep = 0;    // recover initial value
      }
      else
      {
        lst = ACTIONS_SELECTED;
        multiCycleDribbleStep = 1;    // indicate next step
      }
    }
    else // the agent is't fastest to the ball
    {
       soc = SoccerCommand (CMD_ILLEGAL);
       multiCycleDribbleStep = 0;    // recover initial value
       lst = ACTIONS_UNSELECTED_FINISHED;
    }
  }

  if (soc.commandType != CMD_ILLEGAL)
  {
    ACT->putCommandInQueue( soc );
    if (WM->isLastMessageSee())
      ACT->putCommandInQueue(turnNeckDiffSide( soc));
  }


  return lst;

}

/*This method just performs "kick between" with following intercept now.
*/
LevelStateT   Player::outplaySimpleMultiCycle( VecPosition pos )
{
  SoccerCommand soc = SoccerCommand (CMD_ILLEGAL);
  LevelStateT   lst;
  //double d1;

  if (!WM->isConfidenceGood(OBJECT_BALL))
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
    return lst;
  }

  if ( WM->isBallKickable() )
  {
    if ( multiCycleDribbleStep == 0 ) // first stage of the outplaying
    {
      soc = kickTo( pos, SS->getBallSpeedMax()/2.0);
      {
        lst = ACTIONS_SELECTED;
      }
    }
    else // the ball is kickable after an interception, the dribble is finished succesfully
    {
      if (soc.commandType != CMD_ILLEGAL)
      {
          ACT->putCommandInQueue( soc );
          ACT->putCommandInQueue(turnNeckDiffSide( soc));
      }
      multiCycleDribbleStep = 0;    // recover initial value
      lst = ACTIONS_UNSELECTED_FINISHED;
    }
  }
  else  // the ball is not kickable
  {
    // if the agent is still fastest to the ball
    if (WM->getFastestInSetTo(OBJECT_SET_TEAMMATES, OBJECT_BALL) == WM->getAgentObjectType())
    {
      soc = intercept(false, INTERCEPTION_AS_SECOND);
      if (soc.commandType == CMD_ILLEGAL)        // action failed
      {
        lst = ACTIONS_UNSELECTED_FAILED;
        multiCycleDribbleStep = 0;    // recover initial value
      }
      else
      {
        lst = ACTIONS_SELECTED;
        multiCycleDribbleStep = 1;    // indicate next step
      }
    }
    else // the agent is't fastest to the ball
    {
       soc = SoccerCommand (CMD_ILLEGAL);
       multiCycleDribbleStep = 0;    // recover initial value
       lst = ACTIONS_UNSELECTED_FINISHED;
    }
  }

  if (soc.commandType != CMD_ILLEGAL)
  {
    ACT->putCommandInQueue( soc );
    if (WM->isLastMessageSee())
      ACT->putCommandInQueue(turnNeckDiffSide( soc));
  }
  // Vadim begin
  // the SayBall message with these parameters will only be sent
  // when the player kicks the ball and has seen it
  WM->iSayBallMsgCode     = WM->MSG_I_DRIBBLE;
  WM->iNrCyclesToComplete = 5;                  // this is a DUMMY value
  WM->iPlayerInvolved     = WM->getPlayerNumber();
// Vadim end


  return lst;
}


// \param o player to receive this pass
// \param passType kind of pass (either PASS_NORMAL or PASS_FAST )
LevelStateT Player::directPassMultiCycle( ObjectT o, PassT passType   )
{
  int tmNr = SoccerTypes::getTeammateNumber( o );  // pass-receiving teammate number
  // send the ball in the player position predicted by magic nCyc cycles
  int nCyc = 2;
  VecPosition pos = WM->predictPlayerPos( nCyc, o );
  return directPassMultiCycle( pos, tmNr, passType );
}


//! This is a modification of directPassMultiCycle() method for passing to pre-calculated point.
//    \param pos is the point to where the direct pass should be given
//    \param tmNr is the teammate number to whom this pass is directed
//    \param passType kind of pass (either PASS_NORMAL or PASS_FAST )
//    \return SoccerCommand to perform a direct pass to pos           VK***
LevelStateT Player::directPassMultiCycle(  VecPosition pos, int tmNum, PassT passType )
{
  SoccerCommand soc;
  LevelStateT   lst;

  if ( !WM->isConfidenceGood(OBJECT_BALL) )
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
    // drop the idea of passing multi-cycle, if any
    WM->multiCycleAction = ACT_ILLEGAL;
    return lst;
  }

  if ( WM->isBallKickable() )
  {
    soc = directPass( pos, tmNum, passType );
    lst = ACTIONS_SELECTED;
/*
    if ( WM->multiCycleAction == ACT_PASS_ON_REQUEST )
      cout << "#" << WM->getPlayerNumber() << "-->"
          << tmNum << " requested direct-multi to "
          << pos
          << " time = " << WM->getCurrentCycle()
          << endl;
*/          
  }
  else
  {
    soc = SoccerCommand ( CMD_ILLEGAL );
    lst = ACTIONS_UNSELECTED_FINISHED;
    // drop the idea of passing multi-cycle, if any
    WM->multiCycleAction = ACT_ILLEGAL;
  }

  if ( soc.commandType != CMD_ILLEGAL )
  {
    ACT->putCommandInQueue( soc );
    ACT->putCommandInQueue( turnNeckDiffSide(soc) );
  }

  return lst;

} // directPassMultiCycle

LevelStateT Player::leadingPassMultiCycle(  ObjectT o )
{
  SoccerCommand soc;
  LevelStateT   lst;

  if ( !WM->isConfidenceGood(OBJECT_BALL) )
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
    // drop the idea of passing multi-cycle, if any
    WM->multiCycleAction = ACT_ILLEGAL;
    return lst;
  }

  if ( WM->isBallKickable() )
  {
    soc = leadingPass( o, 3.0 );
    lst = ACTIONS_SELECTED;
  }
  else
  {
    soc = SoccerCommand ( CMD_ILLEGAL );
    lst = ACTIONS_UNSELECTED_FINISHED;
    // drop the idea of passing multi-cycle, if any
    WM->multiCycleAction = ACT_ILLEGAL;
  }

  if ( soc.commandType != CMD_ILLEGAL )
  {
    ACT->putCommandInQueue( soc );
    ACT->putCommandInQueue( turnNeckDiffSide(soc) );
  }

  return lst;

} // leadingPassMultiCycle


// hide the ball
/* This method allowes the agent to kick the ball close to the body in the
opposite direction from the closest opponent
*/
LevelStateT   Player::hideBallMultiCycle(ObjectT o)
{
  SoccerCommand soc;
  LevelStateT   lst;
  AngDeg angOpp  = WM->getRelativeAngle(o, true);
  AngDeg angBall = WM->getRelativeAngle(OBJECT_BALL, true);
  AngDeg angDiff = fabs( VecPosition::normalizeAngle(angOpp - angBall) );

  multiCycleDribbleStep = nextMultyCycleCount();
  if ( multiCycleDribbleStep > 3 )
  {
    multiCycleDribbleStep = 0;
    lst =  ACTIONS_UNSELECTED_FINISHED;
    return lst;
  }
  // conditions for exit from the method
  if ( !WM->isConfidenceGood( OBJECT_BALL ) )
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
    return lst;
  }
  
  if ( !WM->isBallKickable() )
  {
    return  ACTIONS_UNSELECTED_FINISHED;
  }

  
  if ( !angDiff < 90.0)
  {
    angOpp = VecPosition::normalizeAngle( angOpp + 180.0);
    soc = kickBallCloseToBody( angOpp );
    lst = ACTIONS_SELECTED;
/*
    cout << "#" << WM->getPlayerNumber()
         << ", hiding ball from opponent #" << SoccerTypes::getOpponentNumber( o )
         << ", cycle = " << WM->getCurrentCycle()
         << endl;
*/
  }
  else
  {
    soc = SoccerCommand (CMD_ILLEGAL);
    lst = ACTIONS_UNSELECTED_FINISHED;
  }

  if (soc.commandType != CMD_ILLEGAL)
  {
    ACT->putCommandInQueue( soc );
    ACT->putCommandInQueue(turnNeckDiffSide(soc));
  }

  return lst;

}  



// Serguei. This method defines if it is break in and the goalie (agent) must intercept
bool  Player::isItBreakIn(ObjectT teammate, ObjectT opponent, VecPosition posBall)
{

   // AngDeg leftAngle =(VecPosition(0.0, -PITCH_WIDTH*0.5) - WM->getPosOwnGoal()).getDirection();
    //AngDeg rightAngle =(VecPosition(0.0, -PITCH_WIDTH*0.5) - WM->getPosOwnGoal()).getDirection();
    AngDeg ballDir  = (WM->getPosOwnGoal() - posBall ).getDirection();
   // AngDeg dirBall  = ( posBall - WM->getPosOwnGoal() ).getDirection();
    AngDeg teamDir  = (WM->getGlobalPosition(teammate) - posBall).getDirection();
    ObjectT secOpponent = WM->getSecondClosestInSetTo(OBJECT_SET_OPPONENTS, OBJECT_BALL);
    bool breakIn = false;

    if ( fabs(posBall.getY()) < 10.0 && WM->isInDefenseArea(OBJECT_BALL) )
    {
      if(WM->getGlobalPosition(secOpponent).getDistanceTo(WM->getPosOwnGoal()) >
         WM->getGlobalPosition(opponent).getDistanceTo(WM->getPosOwnGoal()) )
      {
        if(teammate != opponent)
        {
          if( teammate == WM->getAgentObjectType()  ||
            fabs(ballDir - teamDir) > 120.0)

          {
            breakIn = true;
          }

        }
      }
    }
    return breakIn;

}

/* if level 2 gives a command intercept the player will try to intercept until
success or fail*/

LevelStateT Player::interceptSimpleMultiCycle ( bool isGoalie )
{
  SoccerCommand soc = SoccerCommand(CMD_ILLEGAL);
  LevelStateT   lst;
  InterceptionT iType = INTERCEPTION_AS_SECOND;

  if ( !WM->isConfidenceGood( OBJECT_BALL ) )
  {
    lst =  ACTIONS_UNSELECTED_FINISHED;
    return lst;
  }

  if ( WM->isBallKickable() )    // intercept finished successfully
  {
    soc = SoccerCommand (CMD_ILLEGAL);
    lst = ACTIONS_UNSELECTED_FINISHED;
    //cout<<" finished succs"<<endl;
  }

  else   // an interception is in the process

  {

   if( shouldIIntercept( &iType ) )
   {
       soc = intercept( isGoalie, iType );

      if ( soc.commandType == CMD_ILLEGAL)        // action failed
      {
        lst = ACTIONS_UNSELECTED_FAILED;
       // cout<<" inrecept returned CMD_ILLEGAL"<<endl;
      }
      else
      {
       lst = ACTIONS_SELECTED;
      }
    }
    else // I am not fastest, intercept failed
    {
     soc = SoccerCommand (CMD_ILLEGAL);
     lst = ACTIONS_UNSELECTED_FAILED;
     //cout<<" I am not fastest";
    }
  }

  if ( soc.commandType != CMD_ILLEGAL )
  {
    ACT->putCommandInQueue( soc );
    if ( WM->isLastMessageSee() )
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
  }
  return lst;
}

LevelStateT Player::zoningDefenseMultiCycle(VecPosition center, bool isFree )
{
  //LevelStateT lst; // variable to return
   SoccerCommand action;
   Rectangle defRect = Rectangle(VecPosition(center.getX()-6.5,center.getY()-6.5),
                                 VecPosition(center.getX()+6.5,center.getY()+6.5))  ;
   VecPosition posBall = WM->getBallPos();
  // check conditions to continue method or not
  // I can intercept or  general conditions are changed
  if ((WM->getFastestInSetTo(OBJECT_SET_TEAMMATES, OBJECT_BALL) ==
            WM->getAgentObjectType() && intercept(false, INTERCEPTION_AS_SECOND).commandType!=CMD_ILLEGAL) ||
            (WM->getBallPos().getX() >= -PITCH_LENGTH/6.0 ) ||
            WM->getBallPossesion() == POSSESION_US)

            //(!WM->isConfidenceGood(OBJECT_BALL)) ||
  {
    return ACTIONS_UNSELECTED_FINISHED;
  }


  // defense in progress
  // calculate the point to stay  and point to see

  VecPosition posToSee = center + (posBall-center)/((posBall-center).getMagnitude()*0.5);
  VecPosition posToStay;
  if (posBall.getY() > center.getY())  // the ball is on the right
    posToStay = posToSee+(posBall-center).rotate(90)/((posBall-center).getMagnitude()*0.33);
  else
    posToStay = posToSee+(posBall-center).rotate(-90)/((posBall-center).getMagnitude()*0.33);

  ObjectT closeOpponent = WM->getClosestInSetTo(OBJECT_SET_OPPONENTS, center);
  if (WM->getAgentGlobalPosition().getDistanceTo( posToStay ) > 1.0 )
  {
    action = moveToPos(posToStay, 20, 2);
    ACT->putCommandInQueue( action );
    ACT->putCommandInQueue(turnNeckToObject(OBJECT_BALL, action));
    return ACTIONS_SELECTED;
  }
  else if (closeOpponent!=OBJECT_ILLEGAL &&
     defRect.isInside(WM->getGlobalPosition(closeOpponent)))
  {
    action = mark(closeOpponent, 2.0, MARK_BISECTOR);
    ACT->putCommandInQueue( action );
    return ACTIONS_SELECTED;
  }
  else // we are right in the point
  {
    action =turnBodyToPoint(posToSee,1);
    ACT->putCommandInQueue( action );
    ACT->putCommandInQueue(turnNeckToObject(OBJECT_BALL, action));
    return ACTIONS_SELECTED;
  }


}

LevelStateT   Player::zblokOpponentMultiCycle  ( ObjectT o, double dist, double interval)
{
  //LevelStateT lst; // variable to return
   SoccerCommand action;

   VecPosition posOpp = WM->getGlobalPosition(o);


   VecPosition  posGoal = WM->getPosOwnGoal();
   VecPosition  posToStay= posOpp + (posGoal-posOpp)/((posGoal-posOpp).getMagnitude())*dist;

  // check conditions to continue method or not
  // I can intercept or  general conditions are changed
  if (WM->getFastestInSetTo(OBJECT_SET_TEAMMATES, OBJECT_BALL) !=
            WM->getAgentObjectType() ||
            WM->getBallPossesion() == POSSESION_US ||
            posOpp.getDistanceTo(WM->getAgentGlobalPosition()) > 5.0)
                        
  {
    
    return ACTIONS_UNSELECTED_FINISHED;
  }


  // bloking in progress
  // calculate the point to stay  and point to see

  
  
  
  if (WM->getAgentGlobalPosition().getDistanceTo( posToStay ) > interval )
  {
    //cout<<posOpp<<"  "<<posToStay<<endl;
    action = moveToPos(posToStay, 30, 3);
    ACT->putCommandInQueue( action );
    ACT->putCommandInQueue(turnNeckToObject(o, action));
    return ACTIONS_SELECTED;
  }
  else 
  {
    action =turnBodyToObject(OBJECT_BALL);
    ACT->putCommandInQueue( action );
    ACT->putCommandInQueue(turnNeckToObject(OBJECT_BALL, action));
    return ACTIONS_SELECTED;
  }
  
}

// actually not multicycle now, but will develop later
LevelStateT Player::anticipateBallMultiCycle  ( double dist )
{
  SoccerCommand action;
  VecPosition posAgent = WM->getAgentGlobalPosition();
  VecPosition posBall = WM->getBallPos();
  VecPosition posOpponent = WM->getGlobalPosition(WM->getClosestInSetTo(OBJECT_SET_OPPONENTS, posAgent));
  VecPosition posToStay;
  if (posAgent.getDistanceTo(posOpponent) > dist)
  {
    //cout<<"no need\n";
    action = turnBodyToObject(OBJECT_BALL);
    ACT->putCommandInQueue( action );
    ACT->putCommandInQueue(turnNeckDiffSide(action));
    return ACTIONS_SELECTED_FINISHED;
  }
  
  if (posBall.getDistanceTo(posAgent) < posBall.getDistanceTo(posOpponent))
  {
   //cout<<"behind\n";
    posToStay = posAgent+VecPosition(dist, (posAgent-posOpponent).getDirection(), POLAR);
  }
  else if( (posOpponent-posAgent).getDirection() > (posBall-posAgent).getDirection())
  {
    //cout<<"right\n";
    posToStay = posAgent+VecPosition(dist, (posOpponent-posAgent).getDirection()-90.0, POLAR);
  }
  else
  {
    //cout<<"left\n";
    posToStay = posAgent+VecPosition(dist, (posOpponent-posAgent).getDirection()+90.0, POLAR);
  }

    action = moveToPos(posToStay, 30, 3);
    ACT->putCommandInQueue( action );
    ACT->putCommandInQueue(turnNeckToObject(OBJECT_BALL, action));
    return ACTIONS_SELECTED_FINISHED;
}




// sets the variable indicates the agent possession the ball
void Player::setPlayerPossesion( )
{
  ifIPossessBall = WM->isBallKickable();
  return;
} 


/* this method sets current acton for level 1 */
void Player::setCurrentAction()
{
  currentAction = determineActionMode1();

  if (currentAction == ACT_KICK_BALL)
  {
    if ( WM->multiCycleAction == ACT_PASS_ON_REQUEST )
    {
      currentAction = ACT_DIRECT_PASS;
    }
    else 
    {
      // execute anything else with the ball
      currentAction = determineActionWithBall1();
    }

  }  

  return;
} 



/* the general idea is that we have some directions and some teammates to collaborate
The sequence of wich actions are considered must be the same for any region, but directions and
teammates must be different
*/
ActionT Player::determineActionWithBall1()
{

  ActionT action = ACT_ILLEGAL;
  int i;  //an index for loops

  int tIndex = 10; //number of possible teammates for collaboration. Temporary here, must be defined somewere else
  int dIndex = 6; //number of possible directions for diff purposes. Temporary here, must be defined somewere else


  // general positios
  VecPosition posBall = WM->getBallPos();
  VecPosition bestScoringPoint;

  // general objects
  ObjectT     closestOpponent;
  ObjectT     agentObjectType = WM->getAgentObjectType();

  // the list of free teammates
  ObjectT     leftBackTeammate;
  ObjectT     leftTeammate;
  ObjectT     leftForwardTeammate;
  ObjectT     forwardTeammate;
  ObjectT     rightForwardTeammate;
  ObjectT     rightTeammate;
  ObjectT     rightBackTeammate;
  ObjectT     backTeammate;
  ObjectT     bestTeammateInArea6;
  ObjectT     bestTeammateInArea6a;

  // teammetes for ordering for each region
  ObjectT     tMates[tIndex];
  for( i=0; i<tIndex; i++)
    tMates[i] = OBJECT_ILLEGAL;

  // the list of general directions and angles
  AngDeg      widestForwardDirection = WM->getBestDirForDribble(0.0, 105.0);
  AngDeg      widestStrightForwardDirection = WM->getBestDirForDribble(0.0, 60.0);
  AngDeg      widestBackDirection = WM->getBestDirForDribble((VecPosition(0.0,0.0)-posBall).getDirection(), 60.0);
  AngDeg      towardsGoalDirection = (WM->getPosOpponentGoal()-posBall).getDirection();
  AngDeg      towardsPenaltyAreaDirection = (VecPosition(PITCH_LENGTH/2.0-11.0, 0.0)-posBall).getDirection();


  // directions for ordering
  AngDeg     directions[dIndex];
  for( i=0; i<dIndex; i++)
    directions[i] = 1000.0;

  AngDeg      toTurnDirection = 1000.0;
  AngDeg      leftStickRelDir = (VecPosition(PITCH_LENGTH/2.0, -8.5) - posBall).getDirection();
  AngDeg      rightStickRelDir = (VecPosition(PITCH_LENGTH/2.0, 8.5) - posBall).getDirection();

  double      closestOpponentDistance = 0;

  mapFreeTeammates(&leftBackTeammate, &leftTeammate, &leftForwardTeammate, &forwardTeammate,
                   &rightForwardTeammate, &rightTeammate, &rightBackTeammate,
                   &backTeammate, &bestTeammateInArea6, &bestTeammateInArea6a,
                   posBall);


  closestOpponent = WM->getClosestInSetTo(OBJECT_SET_OPPONENTS, agentObjectType, &closestOpponentDistance);

  towardsGoalDirection = WM->getBestDirForDribble(towardsGoalDirection, 60.0);

  towardsPenaltyAreaDirection = WM->getBestDirForDribble(towardsPenaltyAreaDirection, 60.0);
  leftStickRelDir = WM->getBestDirForDribble(leftStickRelDir, 60.0);
  rightStickRelDir = WM->getBestDirForDribble(rightStickRelDir, 60.0);
  /*if ( goalieRelDir - leftStickRelDir > rightStickRelDir - goalieRelDir )
  {
    scoringMaxAngle = goalieRelDir - leftStickRelDir;
    scoringDirection = leftStickRelDir+1.0;
  }
  else
  {
    scoringMaxAngle = rightStickRelDir - goalieRelDir;
    scoringDirection = rightStickRelDir-1.0;
  }*/

  /* !!! This is the section to ordering teammates and directions according
     regions of thr ball
  */
  if (WM->isInDefenseArea(OBJECT_BALL))
  {
    // teammates
    tMates[1] = forwardTeammate;
    if (posBall.getY() > 0.0)
    {
      tMates[2] = rightForwardTeammate;
      tMates[3] = leftForwardTeammate;
      tMates[5] = rightTeammate;
      tMates[6] = leftTeammate;
    }
    else
    {
      tMates[3] = rightForwardTeammate;
      tMates[4] = leftForwardTeammate;
      tMates[6] = rightTeammate;
      tMates[5] = leftTeammate;
    }

    // don't pass too close to own goal
    for(i=0; i< tIndex; i++)
    {
      if( WM->getGlobalPosition(tMates[i]).getDistanceTo(WM->getPosOwnGoal()) < 20.0)
      {
        tMates[i] = OBJECT_ILLEGAL;
      }
    }

    //directions
    directions[0] = widestStrightForwardDirection;
    directions[1] = towardsGoalDirection;
    directions[2] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    directions[4] = widestBackDirection;
    if (posBall.getY() > 0.0)
      directions[5] = (VecPosition(0.0, PITCH_WIDTH*0.5) - posBall).getDirection();
    else
      directions[5] = (VecPosition(0.0, -PITCH_WIDTH*0.5) - posBall).getDirection();
    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);

    toTurnDirection = 0.0;


  }
  else if (WM->isInMiddleArea(OBJECT_BALL))
  {
    tMates[1] = forwardTeammate;
    if (posBall.getY() > 0.0)
    {
      tMates[2] = rightForwardTeammate;
      tMates[3] = leftForwardTeammate;
      tMates[5] = rightTeammate;
      tMates[6] = leftTeammate;
      tMates[7] = rightBackTeammate;
      tMates[8] = leftBackTeammate;
    }
    else
    {
      tMates[3] = rightForwardTeammate;
      tMates[2] = leftForwardTeammate;
      tMates[6] = rightTeammate;
      tMates[5] = leftTeammate;
      tMates[8] = rightBackTeammate;
      tMates[7] = leftBackTeammate;
    }
     //directions
    directions[0] = widestStrightForwardDirection;
    directions[1] = towardsGoalDirection;
    directions[2] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    toTurnDirection = 0.0;
    directions[4] = widestBackDirection;
    if(posBall.getX() < 0.0)
    {
      if (posBall.getY() > 0.0)
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, PITCH_WIDTH*0.5) - posBall).getDirection();
      else
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, -PITCH_WIDTH*0.5) - posBall).getDirection();

    }
    else
    {
      if (posBall.getY() > 0.0)
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, -PITCH_WIDTH*0.5) - posBall).getDirection();
      else
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, PITCH_WIDTH*0.5) - posBall).getDirection();
    }
    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);

    toTurnDirection = 0.0;
  }
  else if (WM->isBallWithinRegion(REGION_4a) )
  {
    tMates[0] = bestTeammateInArea6a;
    tMates[1] = bestTeammateInArea6;
    tMates[2] = forwardTeammate;
    tMates[3] = leftForwardTeammate;
    tMates[4] = rightForwardTeammate;
    tMates[5] = leftTeammate;
    tMates[6] = rightTeammate;
    tMates[7] = leftBackTeammate;
    tMates[8] = rightBackTeammate;
    tMates[9] = backTeammate;

     //directions

    directions[0] = towardsGoalDirection;
    directions[2] = widestStrightForwardDirection;
    directions[1] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    toTurnDirection = 0.0;
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    directions[5] = (VecPosition(PITCH_LENGTH*0.5, -PITCH_WIDTH*0.5) - posBall).getDirection();
    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);
  }
  else if (WM->isBallWithinRegion(REGION_4b) )
  {
    tMates[0] = bestTeammateInArea6a;
    tMates[1] = bestTeammateInArea6;
    tMates[2] = forwardTeammate;
    tMates[4] = leftForwardTeammate;
    tMates[3] = rightForwardTeammate;
    tMates[6] = leftTeammate;
    tMates[5] = rightTeammate;
    tMates[8] = leftBackTeammate;
    tMates[7] = rightBackTeammate;
    tMates[9] = backTeammate;

    directions[0] = towardsGoalDirection;
    directions[2] = widestStrightForwardDirection;
    directions[1] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    toTurnDirection = towardsGoalDirection;
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    directions[5] = (VecPosition(PITCH_LENGTH*0.5, PITCH_WIDTH*0.5) - posBall).getDirection();
    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);
  }
  else if (WM->isBallWithinRegion(REGION_5a) )
  {
    tMates[0] = bestTeammateInArea6a;
    tMates[1] = bestTeammateInArea6;
    tMates[2] = leftForwardTeammate;
    tMates[3] = leftTeammate;
    tMates[4] = leftBackTeammate;
    tMates[5] = backTeammate;

    //directions
    //directions[0] = towardsPenaltyAreaDirection;
    directions[0] = leftStickRelDir;
    directions[1] = towardsGoalDirection;
    directions[2] = towardsPenaltyAreaDirection;
    directions[3] = (VecPosition(PITCH_LENGTH/2.0-25.0, 0.0)-posBall).getDirection();
    directions[3] = WM->getBestDirForDribble(directions[2], 20.0);
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    directions[5] = directions[0];


   // if (posBall.getX() < PITCH_LENGTH/2.0 - 7.0)
    //  directions[3] = widestForwardDirection;

    toTurnDirection = towardsPenaltyAreaDirection;
   // widestBackDirection = WM->getBestDirForDribble(179.0, 20.0);
  }
  else if (WM->isBallWithinRegion(REGION_5b) )
  {
    tMates[0] = bestTeammateInArea6a;
    tMates[1] = bestTeammateInArea6;
    tMates[2] = rightForwardTeammate;
    tMates[3] = rightTeammate;
    tMates[4] = rightBackTeammate;
    tMates[5] = backTeammate;

    //directions
   // directions[0] = towardsPenaltyAreaDirection;
    directions[0] = rightStickRelDir;
    directions[1] = towardsGoalDirection;
    directions[2] = towardsPenaltyAreaDirection;
    directions[3] = (VecPosition(PITCH_LENGTH/2.0-25.0, 0.0)-posBall).getDirection();
    directions[3] = WM->getBestDirForDribble(directions[2], 20.0);
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    //if (posBall.getX() < PITCH_LENGTH/2.0 - 7.0)
     // directions[3] = widestForwardDirection;
    toTurnDirection = towardsPenaltyAreaDirection;
    //widestBackDirection = WM->getBestDirForDribble(179.0, 20.0);
    directions[5] = directions[0];
  }
  else if (WM->isBallWithinRegion(REGION_6) )
  {
    tMates[1] = forwardTeammate;
    if (posBall.getY() < 0.0)
    {
      tMates[2] = rightForwardTeammate;
      tMates[3] = leftForwardTeammate;
      tMates[5] = rightTeammate;
      tMates[6] = leftTeammate;
      tMates[7] = rightBackTeammate;
      tMates[8] = leftBackTeammate;
    }
    else
    {
      tMates[3] = rightForwardTeammate;
      tMates[2] = leftForwardTeammate;
      tMates[6] = rightTeammate;
      tMates[5] = leftTeammate;
      tMates[8] = rightBackTeammate;
      tMates[7] = leftBackTeammate;
    }
    directions[2] = widestStrightForwardDirection;
    directions[0] = towardsGoalDirection;
    directions[1] = towardsPenaltyAreaDirection;
    directions[3] = widestForwardDirection;
    directions[4] = WM->getBestDirForDribble((VecPosition(24.0, 0.0)-posBall).getDirection(), 20.0);
    toTurnDirection = 0.0;
    if (posBall.getY() > 0.0)
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, -PITCH_WIDTH*0.5) - posBall).getDirection();
      else
        directions[5] = (VecPosition(PITCH_LENGTH*0.5, PITCH_WIDTH*0.5) - posBall).getDirection();

    directions[5] = WM->getBestDirForDribble(directions[5] , 30.0);

  }
   else if (WM->isBallWithinRegion(REGION_6a) )
  {
    tMates[0] = bestTeammateInArea6a;
    if (posBall.getY() > 0.0)
    {
      tMates[1] = bestTeammateInArea6;
      tMates[2] = leftTeammate;
      tMates[3] = rightTeammate;
      tMates[4] = leftBackTeammate;
      tMates[5] = rightBackTeammate;
    }
    else
    {
      tMates[1] = bestTeammateInArea6;
      tMates[3] = leftTeammate;
      tMates[2] = rightTeammate;
      tMates[5] = leftBackTeammate;
      tMates[4] = rightBackTeammate;
    }
    tMates[6] = backTeammate;

    directions[0] = towardsPenaltyAreaDirection;
    directions[1] = towardsGoalDirection;
    directions[4] = widestBackDirection;
    directions[5] = directions[0];

    toTurnDirection = towardsPenaltyAreaDirection;
  }


  if (WM->isDeadBallUs())
  {
     if(WM->getTimeSinceLastCatch() < PS->getCyclesCatchWait() ||
        WM->isGoalKickUs() )
     {
        action = ACT_MOVE_TO_DEAD_BALL_POSITION;
     }

     // for right kicking the agent must go to ridght position and turn at first
     if (WM->getAgentGlobalPosition().getDistanceTo(WM->getPosForDeadBall()) > 0.5 ||
          WM->getRelativeAngle(OBJECT_BALL, true) > 15.0 )
     {
       if (WM->getAgentViewAngle() !=VA_WIDE)
       {
         ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_WIDE, VQ_HIGH ) );
       }
       action = ACT_MOVE_TO_DEAD_BALL_POSITION;
     }

     // shooting
      if ( action == ACT_ILLEGAL && shouldIScore( &bestScoringPoint ) )
      {
          action = ACT_KICK_TO;
          pointToPass = bestScoringPoint;
      }
      // end of shooting

      if (action == ACT_ILLEGAL) // action wasn't chosen befor
      //trying to give a leading pass
      {
        for(i=0; i < 4; i++) // empiric number here, must be defined somewhere else
        {
          if (tMates[i]!= OBJECT_ILLEGAL && WM->canIGiveLeadingPass(tMates[i], &pointToPass)  )
          {
             //cout<<" leading pass "<<i<<endl;
            //pointToPass = getLeadingPassPoint(tMates[i]);
            action = ACT_LEADING_PASS;
            teammateToPass = tMates[i];
            break;
          }
        }
     }

     if (action == ACT_ILLEGAL) // action wasn't chosen
     //trying to find very free teammate
     {
       for(i=0; i < tIndex; i++)
       {
         if (tMates[i]!= OBJECT_ILLEGAL && WM->isVeryFree(tMates[i]))
         {
            action = ACT_DIRECT_PASS;
            teammateToPass = tMates[i];
            break;
         }
       }
     }

     if (action == ACT_ILLEGAL) // action wasn't chosen
     //trying to find fairly  free teammate
     {
       for(i=0; i < tIndex; i++)
       {
         if (tMates[i]!= OBJECT_ILLEGAL && WM->isFairlyFree(tMates[i]))
         {
            action = ACT_DIRECT_PASS;
            teammateToPass = tMates[i];
            break;
         }
       }
     }

    if (action == ACT_ILLEGAL) // action wasn't chosen
     //trying to find just free teammate
     {
       for(i=0; i < tIndex; i++)
       {
         if (tMates[i]!= OBJECT_ILLEGAL && WM->isFree(tMates[i]))
         {
            action = ACT_DIRECT_PASS;
            teammateToPass = tMates[i];
            break;
         }
       }
     }

     if (action == ACT_ILLEGAL) // action is still not chosen
     {
       action = ACT_KICK_TO;
       if(WM->isKickOffUs() )
       {
         pointToPass = WM->getPointBetweenTwoClosestOpponentsInCone(
                      posBall,
                      VecPosition(-PITCH_LENGTH/2.0,-PITCH_WIDTH/2.0),
                      VecPosition(-PITCH_LENGTH/2.0, PITCH_WIDTH/2.0));
       }
       else
       // pointToPass = WM->getPointBetweenTwoClosestOpponentsInCone(
       //               posBall,
       //               VecPosition(PITCH_LENGTH/2.0,-PITCH_WIDTH/2.0),
       //               VecPosition(PITCH_LENGTH/2.0, PITCH_WIDTH/2.0));
        pointToPass = posBall+VecPosition(20.0, WM->getBestDirForDribble(directions[1], 20.0), POLAR);
     }
  } // end of deadBallUs

  // shooting
  else if ( shouldIScore( &bestScoringPoint ) )
  {
    action = ACT_KICK_TO;
    pointToPass = bestScoringPoint;
  }
  // end of shooting

  else if ( closestOpponent != OBJECT_ILLEGAL && shouldITrickOpponent( closestOpponent ) )
  {
    action = ACT_TRICK;
    teammateToPass = closestOpponent;
  }

  else if ( closestOpponent != OBJECT_ILLEGAL && shouldIHideBall( closestOpponent ) )
  {
    action = ACT_HIDE_BALL;
    teammateToPass = closestOpponent;
  }


  /*!< The first part is stright attack. The agent checks 4 teammates and 4 directions
    for the attack
  */
  // leading pass
   if (action == ACT_ILLEGAL) // action wasn't chosen befor
  //trying to find very free teammate to pass with very high confidence
  {
    for(i=0; i < 4; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->canIGiveLeadingPass(tMates[i], &pointToPass)  )
       //&&   WM->getConfidence(tMates[i]) > PS->getPlayerHighConfThr())
      {
         //cout<<" leading pass "<<i<<endl;
         //pointToPass = getLeadingPassPoint(tMates[i]);
         action = ACT_LEADING_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }

  if (action == ACT_ILLEGAL) // action wasn't chosen before
  //trying to find very free teammate to pass with very high confidence
  {
    for(i=0; i < 4; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->isVeryFree(tMates[i])  )
       //&&   WM->getConfidence(tMates[i]) > PS->getPlayerHighConfThr())
      {
        // cout<<"pass teammate  very free"<<i<<endl;
         action = ACT_DIRECT_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }

  if (action == ACT_ILLEGAL) // action wasn't chosen before
  //trying to find very free teammate to pass with very high confidence
  {
    for(i=0; i < 4; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->isFairlyFree(tMates[i])  )
       //&&   WM->getConfidence(tMates[i]) > PS->getPlayerHighConfThr())
      {
        // cout<<"pass teammate  very free"<<i<<endl;
         action = ACT_DIRECT_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }
  // leading pass
   if (action == ACT_ILLEGAL) // action wasn't chosen befor
  //trying to find very free teammate to pass with very high confidence
  {
    for(i=4; i < 6; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->canIGiveLeadingPass(tMates[i], &pointToPass)  )
       //&&   WM->getConfidence(tMates[i]) > PS->getPlayerHighConfThr())
      {
         //cout<<" leading pass "<<i<<endl;
         pointToPass = getLeadingPassPoint(tMates[i]);
         action = ACT_KICK_TO;
         //teammateToPass = tMates[i];
         break;
      }
    }
  }
  //if need to turn
  if (action == ACT_ILLEGAL) // action wasn't chosen before
  {
    if(WM->canITurn(toTurnDirection))
    {
      //cout<<"turn "<<endl;
      action = ACT_TURN_WITH_BALL;
      dribbleDirection = toTurnDirection;
    }
  }

  // dribble fast
  if (action == ACT_ILLEGAL) // action wasn't chosen before
  {
    for(i=0; i < 3; i++) // empiric number here, must be defined somewhere else
    {
      if(directions[i] < 900.0 && WM->canItoDribble( directions[i], DRIBBLE_FAST))
      {
        //cout<<"dribble fast"<<endl;
        action = ACT_DRIBBLE;
        dribbleDirection = directions[i];
        pointToPass = posBall;
        dribbleType = DRIBBLE_FAST;
        break;
      }
    }
  }

  // dribble slow
 if (action == ACT_ILLEGAL) // action wasn't chosen before
  {
    for(i=0; i < 3; i++) // empiric number here, must be defined somewhere else
    {
      if(directions[i] < 900.0 && WM->canItoDribble( directions[i], DRIBBLE_SLOW))
      {
        //cout<<"dribble slow"<<endl;
        action = ACT_DRIBBLE;
        dribbleDirection = directions[i];
        pointToPass = posBall;
        dribbleType = DRIBBLE_SLOW;
        break;
      }
    }
  }



  // fairly free pass

  if (action == ACT_ILLEGAL) // action wasn't chosen befor
  //trying to find fairly free teammate to pass with very high confidence
  {
    for(i=0; i < 4; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->isFree(tMates[i])  )
       //&&   WM->getConfidence(tMates[i]) > PS->getPlayerHighConfThr())
      {
         //cout<<"pass teammate firly free"<<i<<endl;
         action = ACT_DIRECT_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }


  /*!< The second part is support attack. The agent checks 2 teammates
    and outplay for the supporting attack
  */

  if (action == ACT_ILLEGAL) // action wasn't chosen before
  //trying to find very free teammate to pass with very high confidence from second division
  {
    for(i=4; i < 6; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->isVeryFree(tMates[i]))

      {
         //cout<<"pass teammate  very free"<<i<<endl;
         action = ACT_DIRECT_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }

  if (action == ACT_ILLEGAL) // action wasn't chosen before
  //trying to find very free teammate to pass with very high confidence from second division
  {
    for(i=4; i < 6; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->isFairlyFree(tMates[i]))

      {
         //cout<<"pass teammate  very free"<<i<<endl;
         action = ACT_DIRECT_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }


   if (action == ACT_ILLEGAL) // action wasn't chosen before
  //trying to find fairly free teammate
  {
    for(i=4; i < 6; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->isFree(tMates[i]))
      {
         //cout<<"pass teammate "<<i+1<<endl;
         action = ACT_DIRECT_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }

  /*!< The third part is holding the ball. The agent checks 2 teammates
     for the ball holding.
  */
  if (action == ACT_ILLEGAL) // action wasn't chosen before
  //trying to find very free teammate to pass with very high confidence from second division
  {
    for(i=6; i < 8; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->isVeryFree(tMates[i]))

      {
         //cout<<"pass teammate very free"<<i<<endl;
         action = ACT_DIRECT_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }

   /*9if (action == ACT_ILLEGAL && !WM->isInDefenseArea(OBJECT_BALL)) // action wasn't chosen before
  //else
  {
    // temporarily insted outplay
    //cout<<"instead outplay"<<endl;
    if (WM->getAgentViewAngle() !=VA_WIDE)
    {
      ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_WIDE, VQ_HIGH ) );
    }
    action = ACT_HIDE_BALL;
    teammateToPass = closestOpponent;
    */
    //action = ACT_KICK_TO;
    //pointToPass = posBall+VecPosition(20.0, directions[5], POLAR);
    /* for(i=0; i < 3; i++) // empiric number here, must be defined somewhere else
     {
       if(directions[i] < 900.0 && WM->canItoOutplay( directions[i]) )
       {
         // cout<<"outplay!!! "<<endl;
          action = ACT_OUTPLAY;
          pointToPass = posBall+VecPosition(5.0, directions[i], POLAR);
          break;
        }
     }*/
  //}  // end of outplay

  /*
   if (action == ACT_ILLEGAL) // action wasn't chosen before
  //trying to find fairly free teammate
  {
    for(i=6; i < 8; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->isFairlyFree(tMates[i]))
      {
         //cout<<"pass teammate fairly free"<<i<<endl;
         action = ACT_DIRECT_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }
  */

  /*!< The final part is holding the ball ultimately. The agent checks 1 teammate,
     turn back, dribble back and clear the ball for the ball holding.
  */
   if (action == ACT_ILLEGAL) // action wasn't chosen before
  //trying to find fairly free teammate to leading pass with very high confidence
  {
    for(i=8; i < 9; i++) // empiric number here, must be defined somewhere else
    {
      if (tMates[i]!= OBJECT_ILLEGAL && WM->isVeryFree(tMates[i]))
      {
         //cout<<"pass teammate very free"<<i+1<<endl;
         action = ACT_DIRECT_PASS;
         teammateToPass = tMates[i];
         break;
      }
    }
  }

  /*if (action == ACT_ILLEGAL) // action wasn't chosen before
  {
    if(WM->canITurn(directions[4]))
    {
      //cout<<"turn back "<<endl;
      action = ACT_TURN_WITH_BALL;
      dribbleDirection = directions[4];
    }
  }*/


  // dribble back
  if (action == ACT_ILLEGAL&& !WM->isInDefenseArea(OBJECT_BALL)) // action wasn't chosen before
  {

    if(directions[4] < 900.0 && WM->canItoDribble( directions[4], DRIBBLE_WITHBALL))
    {
      //cout<<"dribble back"<<endl;
      action = ACT_DRIBBLE;
      dribbleDirection =directions[4];
      pointToPass = posBall;
      dribbleType = DRIBBLE_WITHBALL;

    }
  }

  if (action == ACT_ILLEGAL && WM->getBallPos().getX() > 0.0) // action wasn't chosen before
  //else
  {
    // temporarily insted outplay
    //cout<<"instead outplay"<<endl;
    if (WM->getAgentViewAngle() !=VA_WIDE)
    {
      ACT->sendCommandDirect( SoccerCommand( CMD_CHANGEVIEW, VA_WIDE, VQ_HIGH ) );
    }
    action = ACT_HIDE_BALL;
    teammateToPass = closestOpponent;


  }
  if (action == ACT_ILLEGAL) // action wasn't chosen before
  //else
  {
       // kick towards the goal
       //cout<<"kick between "<<endl;
       action = ACT_KICK_TO;
       pointToPass = posBall+VecPosition(20.0, directions[0], POLAR);

      // pointToPass =posBall + VecPosition(10.0, t1, POLAR);
      // cout<<posBall<<pointToPass;
       //pointToPass = WM->getPosOpponentGoal();


  }// end

  return action;
}


/*! This method picks the best point in the net to shoot at and returns it
  \return VecPosition holding the best shooting point
  \author David Brokenshire
  ***needs work***
*/
VecPosition Player::selectBestShootingPoint()
{
  //These are the only points we consider shooting at right now.
  static VecPosition targetPoint1 = VecPosition(PITCH_LENGTH/2, -6.0);
  static VecPosition targetPoint2 = VecPosition(PITCH_LENGTH/2, 6.0);
  int numberInFirstCone = 0;
  int numberInSecondCone = 0;
  
  //Need a better method of deciding where to shoot. Make a cone or some other
  //region leading to each of the alternative shooting points, check the nr of
  //opponents in the region, the distance from the ball to the shooting point
  //the distance between the goalie and the ball line.
  numberInFirstCone = WM->getNrInSetInCone(OBJECT_SET_OPPONENTS, 0.25, WM->getBallPos(), targetPoint1);
  numberInSecondCone = WM->getNrInSetInCone(OBJECT_SET_OPPONENTS, 0.25, WM->getBallPos(), targetPoint2);  

  VecPosition goaliePosition = WM->getOpponentGoaliePos();
  Line line1 = Line::makeLineFromTwoPoints(WM->getBallPos(), targetPoint1);
  Line line2 = Line::makeLineFromTwoPoints(WM->getBallPos(), targetPoint2);
  double distanceFromLine1ToGoalie = line1.getDistanceWithPoint(goaliePosition);
  double distanceFromLine2ToGoalie = line2.getDistanceWithPoint(goaliePosition);

  
  if((WM->getBallPos().getDistanceTo(targetPoint1) < 30) && (WM->getBallPos().getDistanceTo(targetPoint2) < 30))
  { if (numberInFirstCone == numberInSecondCone)
    {  if (WM->getAgentGlobalPosition().getY() < 0.0)
        return targetPoint1;
       else
        return targetPoint2;
    }
    else if (numberInFirstCone < numberInSecondCone)
    { return targetPoint1;
    }
    else
    { return targetPoint2;
    }
  }
  else if (WM->getBallPos().getDistanceTo(targetPoint1) < 30)
  { return targetPoint1;
  }
  else if (WM->getBallPos().getDistanceTo(targetPoint2) < 30)
  { return targetPoint2;
  }
  else
  { if (numberInFirstCone == numberInSecondCone)
    {  if (WM->getAgentGlobalPosition().getY() < 0.0)
        return targetPoint1;
       else
        return targetPoint2;
    }
    else if (numberInFirstCone < numberInSecondCone)
    { return targetPoint1;
    }
    else
    { return targetPoint2;
    }
  }
  //The distance between a point and a line.
 
/*
  //stub
  if (WM->getAgentGlobalPosition().getY() < 0.0)
    return VecPosition(PITCH_LENGTH/2, -6.0);
  else
    return pointToPass = VecPosition(PITCH_LENGTH/2, 6.0);
*/    
}
  
