/***************************************************************************
                          level.cpp  -  description
                             -------------------
    begin                : Tue Nov 19 2002
    copyright            : (C) 2002 by David Brokenshire
    email                : brokenshire@sfu.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

 
#include "level.h"

  /*! Basic constructor for Level.  Only called by subclasses
  */
  Level::Level()
  {
    timer = 0;
    higherLevel = NULL;
    lowerLevel = NULL;
  }

  /*! This method updates the timer for the level by one cycle.
      The value will always be >= 0
    /return new value of the timer.
  */
  int Level::updateTimer()
  {  return timer >= 1 ? --timer : timer = 0;
  }

  /*! This method is called to indicate that the level is expired. It sets the level's
      timer to have no time left. (== 0)
  */
  void Level::expireTimer()
  { timer = 0;
  }
  
  /*! This method retrieves a reference to the next lowest level of abstraction in the list.
      This should exist for all levels except the final, for which it should be null
  */
  Level* Level::getLowerLevel()
  { return lowerLevel;
  }

  /*! This method retrieves a reference to the next highest level of abstraction in the list.
      This should exist for all levels except the final, for which it should be null
  */  
  Level* Level::getHigherLevel()
  { return higherLevel;
  }

  /*! This method sets the reference to the next lowest level.
  */
  void Level::setLowerLevel(Level* newLowerLevel)
  { lowerLevel = newLowerLevel;
  }

  /*! This method sets the reference to the next highest level.
  */
  void Level::setHigherLevel(Level* newHigherLevel)
  { higherLevel = newHigherLevel;
  }

  /*! This method gets the current value of the timer.  The timer is >= 0.

      /returns int number of cycles remaining for this level
  */
  int Level::getTimer()
  { return timer;
  }

  /*! This method sets the value of the timer.  The value should be >= 0
      or it will be set to 0.
  */
  void Level::setTimer(int timerValue)
  { timerValue >= 0 ? timer = timerValue : timer = 0;
  }
