/***************************************************************************
                          level.h  -  description
                             -------------------
    begin                : Tue Nov 19 2002
    copyright            : (C) 2002 by David Brokenshire
    email                : brokenshire@sfu.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef LEVEL_H
#define LEVEL_H

#include <vector>
#include "feature.h"

#include "BasicPlayer.h"

class Player;

using namespace std;


/**
  *@author David Brokenshire
  */

/*! This class is an abstract base class from which the decision making architecture of
    the player can be derived.  It provides the basic framework within which to implement
    an arbitrary number of decision making mechanisms and layers of abstraction.  All classes
    which actually implement a procedure for making decisions should be derived from Level.
*/
class Level {

protected:
  //Data
  Level* higherLevel;     /*!<  A reference to the next higher level of abstraction   */
  Level* lowerLevel;      /*!<  A reference to the next lower level of abstraction    */
  int timer;              /*!<  Records how much time the level has left to execute   */
  Player* player;         /*!<  Reference to the player that contains this level      */

  void setTimer(int timerValue);

public: 
	Level();

  /*! A return value of 0 indicates in process completion (actions selected)
      A return value of 1 indicates completion (actions not selected)
      A return value of -1 indicates failure (actions not selected)
  */
  virtual LevelStateT execute() = 0;
//  virtual int execute(vector<Feature> featureVector) = 0;

  int updateTimer();          
  void expireTimer();
  int getTimer();
  Level* getLowerLevel();
  Level* getHigherLevel();
  void setLowerLevel(Level* newLowerLevel);
  void setHigherLevel(Level* newHigherLevel);
  
};

#endif

